from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer, TfidfTransformer
from nltoolkit.helpers import TokenSplitter

from tf_core.nltoolkit.helpers import TokenSplitter


class BowModelConstructor:
    def __init__(self,adc,token_annotation,stem_feature_name,stop_word_feature_name,
                 label_doc_feature_name,
                weighting_type, normalize_vectors,
                 max_ngram,min_tf,predefined_vocabulary=None):

        self._feature_name=token_annotation+('/'+stem_feature_name if stem_feature_name else '')

        self._stop_word_feature_name=stop_word_feature_name
        self._doc_class_label=label_doc_feature_name

        raw_documents=self.get_raw_text(adc.documents,join_annotations_with='|##|')
        #extract vocabulary
        #document_tokens=self.get_annotation_texts(adc.documents)

        # vocabulary={}
        # for raw_doc in document_tokens:
        #     for token in raw_doc:
        #         if not token in vocabulary:
        #             vocabulary[token]=len(vocabulary)

        #(?u)\\b\\w+\\b
        self.__count_params={'ngram_range':(1,max_ngram),'min_df':min_tf,'tokenizer':TokenSplitter(),
                             'token_pattern':'\\b\\w+\\b'} #'vocabulary':vocabulary}
        #tf_idf args: 'norm', 'smooth_idf', 'sublinear_tf', 'use_idf
        self.__tfidf_params=self.__wighting_type_to_tfidf_params(weighting_type)
        self.__tfidf_params['norm']='l2' if normalize_vectors else None

        #set default vectorizerl
        self.vectorizer =self._count_vectorizer() if weighting_type=='term_freq' else self._tfidf_vectorizer() #DictVectorizer(dtype=dtype, sparse=sparse)

        #set predefined controlled vocabulary
        if predefined_vocabulary:
            vocab_vectorizer=CountVectorizer(ngram_range=(1,max_ngram),tokenizer=lambda doc: [doc]) #every line represents a new token
            vocab_vectorizer.fit(predefined_vocabulary)
            self.set_new_vocabulary(vocab_vectorizer.vocabulary_,raw_documents)
        else:
            self.vectorizer.fit(raw_documents) #fit the vectorizer to the documents
            self.__count_params['vocabulary']=self.vectorizer.vocabulary_ #set the learned vocabulary also to future vectorizers

        #print self.vectorizer.get_feature_names()

    def set_new_vocabulary(self,vocabulary,raw_documents,intersect=True):
        if intersect: #intersect vocabularies
            self.vectorizer.fit(raw_documents) #fit the vectorizer to the documents
            feature_intersection=[term for term in self.vectorizer.get_feature_names()
                                  if term in vocabulary]

            self.vectorizer.set_params(vocabulary=dict([(a,i) for i,a in enumerate(feature_intersection)])) #set new vocabulary
            self.vectorizer.fit(raw_documents)
            self.__count_params['vocabulary']=self.vectorizer.vocabulary_ #set the learned vocabulary also to future vectorizers

        else:
            self.__count_params['vocabulary']=vocabulary


    @staticmethod
    def __wighting_type_to_tfidf_params(weighting_type):
        if weighting_type=='term_freq':
            return {}
        elif weighting_type=='tf_idf':
            return {'use_idf':True,'smooth_idf':False,'sublinear_tf':False}
        elif weighting_type=='tf_idf_safe':
            return {'use_idf':True,'smooth_idf':True,'sublinear_tf':False}
        elif weighting_type=='log_df_tf_idf':
            return {'use_idf':True,'smooth_idf':False,'sublinear_tf':True}


    def _vocab_to_idx(self):
        return self.__count_params['vocabulary']


    def _idx_to_vocab(self):
        return {v: k for k, v in self._vocab_to_idx().items()}


    def _count_vectorizer(self):
        return CountVectorizer(**self.__count_params)
    def _tfidf_vectorizer(self):
        return TfidfVectorizer(**dict(self.__count_params.items()
                                      +self.__tfidf_params.items()
                                      #+[['tokenizer',self.custom_tokenizer]]
                                      ))
    def _tfidf_transformer(self):
        return TfidfTransformer(**self.__tfidf_params)



    def get_raw_text(self,documents,join_annotations_with=" "):
        return [document.raw_text(selector=self._feature_name,
                                  join_annotations_with=join_annotations_with,
                                  stop_word_feature_name=self._stop_word_feature_name)
                for document in documents]

    def get_annotation_texts(self,documents):
        return [document.get_annotation_texts(selector=self._feature_name,
                                  stop_word_feature_name=self._stop_word_feature_name)
                for document in documents]


    def get_document_labels(self,adc,binary=False):
        '''
        :param adc: annotated document corpus
        :param binary: return binary results
        :return: for every document if it contains the selected label
        '''

        if self._doc_class_label:
            res=[doc.get_first_label(self._doc_class_label) for doc in adc.documents]
            if binary:
                uniq_res=list(set(res))
                return [uniq_res.index(r) for r in res]
            return res
        else:
            return None
    def get_feature_names(self):
        return self.vectorizer.get_feature_names()

#BowSpace
        # private ITokenizer mTokenizer
        #     = new UnicodeTokenizer();
        # private Set<string>.ReadOnly mStopWords
        #     = null;
        # private IStemmer mStemmer
        #     = null;
        # private Dictionary<string, Word> mWordInfo
        #     = new Dictionary<string, Word>();
        # private ArrayList<Word> mIdxInfo
        #     = new ArrayList<Word>();
        # private int mMaxNGramLen
        #     = 2;
        # private int mMinWordFreq
        #     = 5;
        # private WordWeightType mWordWeightType
        #     = WordWeightType.TermFreq;
        # private double mCutLowWeightsPerc
        #     = 0.2;
        # private bool mNormalizeVectors
        #     = true;
        # private bool mKeepWordForms
        #     = false;



from itertools import izip

from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.tree import DecisionTreeClassifier

from tf_core.nltoolkit.helpers import NltkClassifier


class BowDataset:
    def __init__(self,sparse_bow_matrix,labels=None):
        self.sparse_bow_matrix=sparse_bow_matrix
        self.labels=labels
    def __len__(self):
        return self.sparse_bow_matrix.shape[0]
    @classmethod
    def from_raw_documents(cls,documents,bow_vectorizer,labels=None):
        sparse_bow_matrix = bow_vectorizer.transform(documents)
        return cls(sparse_bow_matrix,labels)

    @classmethod
    def from_adc(cls,adc,bow_model):
        sparse_bow_matrix = bow_model.vectorizer.transform(bow_model.get_raw_text(adc.documents,join_annotations_with='|##|'))
        labels=bow_model.get_document_labels(adc)

        return cls(sparse_bow_matrix,labels)

    #def sparce_bow_matrix(self):
    #    return self.sparse_bow_matrix()
    def dense_bow_matrix(self):
        return self.sparse_bow_matrix.toarray()

    def nltk_dataset_with_labels(self):
        train_data=[({},label) for label in self.labels]
        cx=self.sparse_bow_matrix.tocoo() #A sparse matrix in COOrdinate format.
        if cx.shape[0]!=len(self.labels):
            raise Exception("Nekaj gnilega je v dezeli Danski, sporoci maticu.")
        for (i,j,v) in izip(cx.row, cx.col, cx.data):
            train_data[i][0][j]=v   #seting the dict in (list(tuple(dict, str)))
        return train_data

    def nltk_dataset_without_labels(self):
        cx=self.sparse_bow_matrix.tocoo() #A sparse matrix in COOrdinate format.
        train_data=[{} for _ in range(cx.shape[0])]
        for (i,j,v) in izip(cx.row, cx.col, cx.data):
            train_data[i][j]=v   #seting the dict in (list(dict))
        return train_data

    def bow_in_proper_format(self,classifier,no_labels=False):
        #check if classifier can deal with sparse data
        if isinstance(classifier,NltkClassifier):
            return self.nltk_dataset_without_labels() if no_labels else self.nltk_dataset_with_labels()
        elif isinstance(classifier, (GaussianNB, MultinomialNB,  DecisionTreeClassifier)):
            return self.dense_bow_matrix()
        else: #if latino classifier or a classifier that can deal with sparse data
            return self.sparse_bow_matrix

    def split(self,train_indices,test_indices=None):
        output_train = BowDataset(self.sparse_bow_matrix[train_indices],
                                      self.get_document_labels(train_indices))
        output_test = BowDataset(self.sparse_bow_matrix[test_indices], self.get_document_labels(test_indices)) \
                if test_indices else None
        return output_train,output_test

    def get_document_labels(self,indices):
        return [self.labels[i] for i in indices] if self.labels!=None else []

# try:
#     from nltk.classify import scikitlearn
#     from sklearn.feature_extraction.text import TfidfTransformer
#     from sklearn.pipeline import Pipeline
#     from sklearn import ensemble, feature_selection, linear_model, naive_bayes, neighbors, svm, tree
#
#     classifiers = [
#         ensemble.ExtraTreesClassifier,
#         ensemble.GradientBoostingClassifier,
#         ensemble.RandomForestClassifier,
#         linear_model.LogisticRegression,
#         #linear_model.SGDClassifier, # NOTE: this seems terrible, but could just be the options
#         naive_bayes.BernoulliNB,
#         naive_bayes.GaussianNB,
#         naive_bayes.MultinomialNB,
#         neighbors.KNeighborsClassifier,  # TODO: options for nearest neighbors
#         svm.LinearSVC,
#         svm.NuSVC,
#         svm.SVC,
#         tree.DecisionTreeClassifier,
#     ]



from lib.bag_of_words import *
from lib.classification import *
#from lib.clustering import *
from lib.document_corpus import *
#from lib.similarity_matrix import *
from lib.part_of_speech_tagging import *
from lib.stop_word_removal import *
from lib.stemming import *
from lib.textual_data_in_out import *
from lib.tokenization import *
from lib.chunking import *


from collections import Iterable
def flat(lis):
     for item in lis:
         if isinstance(item, list):# and not isinstance(item, basestring):
             for x in flat(item):
                 yield x
         else:
             yield item

def flatten(lis):
    return list(flat(lis))


from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer

f = open("/root/Myfolder/scoringDocuments/doc1")
doc1 = str.decode(f.read(), "UTF-8", "ignore")
f = open("/root/Myfolder/scoringDocuments/doc2")
doc2 = str.decode(f.read(), "UTF-8", "ignore")
f = open("/root/Myfolder/scoringDocuments/doc3")
doc3 = str.decode(f.read(), "UTF-8", "ignore")

train_set = ["president of India",doc1, doc2, doc3]

tfidf_vectorizer = TfidfVectorizer()
tfidf_matrix_train = tfidf_vectorizer.fit_transform(train_set)  #finds the tfidf score with normalization
print "cosine scores ==> ",cosine_similarity(tfidf_matrix_train[0:1], tfidf_matrix_train)  #here the first element of tfidf_matrix_train is matched with other three elements


from nltk.classify.util import names_demo, names_demo_features



#!/usr/bin/env python

# from nltk.classify.megam import config_megam, call_megam
#from nltk.classify.weka import WekaClassifier, config_weka
from datetime import time
from nltk import ELEProbDist
from nltk.classify.naivebayes import NaiveBayesClassifier
from nltk.classify.positivenaivebayes import PositiveNaiveBayesClassifier
from nltk.classify.decisiontree import DecisionTreeClassifier
from nltk.classify.rte_classify import rte_classifier, rte_features, RTEFeatureExtractor
from nltk.classify.util import accuracy, apply_features, log_likelihood
from nltk.classify.scikitlearn import SklearnClassifier
from nltk.classify.maxent import (MaxentClassifier, BinaryMaxentFeatureEncoding,
                                  TypedMaxentFeatureEncoding,
                                  ConditionalExponentialClassifier)
#################
#OBSERVATIONS
#################
#GaussianNB is not a good fit for document classification at all, since tf-idf values are non-negative frequencies;
# use MultinomialNB instead, and maybe try BernoulliNB. scikit-learn comes with a document classification example that,
# incidentally, uses tf-idf weighting using the built-in TfidfTransformer.
#################
from tf_core.nltoolkit.helpers import NltkClassifier, DictionaryProbDist


def nltk_naive_bayes_classifier(input_dict):
    """
    A classifier based on the Naive Bayes algorithm.  In order to find the
    probability for a label, this algorithm first uses the Bayes rule to
    express P(label|features) in terms of P(label) and P(features|label):

    |                       P(label) * P(features|label)
    |  P(label|features) = ------------------------------
    |                              P(features)

    The algorithm then makes the 'naive' assumption that all features are
    independent, given the label:

    |                       P(label) * P(f1|label) * ... * P(fn|label)
    |  P(label|features) = --------------------------------------------
    |                                         P(features)

    Rather than computing P(featues) explicitly, the algorithm just
    calculates the denominator for each label, and normalizes them so they
    sum to one:

    |                       P(label) * P(f1|label) * ... * P(fn|label)
    |  P(label|features) = --------------------------------------------
    |                        SUM[l]( P(l) * P(f1|l) * ... * P(fn|l) )


    """
    estimator = ELEProbDist #TODO estimator
    classifier=NltkClassifier(NaiveBayesClassifier,estimator=estimator)

    return {'classifier': classifier}


from nltk.classify import DecisionTreeClassifier, MaxentClassifier, NaiveBayesClassifier, megam
#from nltk_trainer import basestring
#from nltk_trainer.classification.multi import AvgProbClassifier

#
# classifier_choices = ['NaiveBayes', 'DecisionTree', 'Maxent'] + MaxentClassifier.ALGORITHMS
#
# dense_classifiers = set(['ExtraTreesClassifier', 'GradientBoostingClassifier',
#                          'RandomForestClassifier', 'GaussianNB', 'DecisionTreeClassifier'])
# verbose_classifiers = set(['RandomForestClassifier', 'SVC'])
#
# try:
#     import svmlight  # do this first since svm module makes ugly errors
#     from nltk.classify.svm import SvmClassifier
#
#     classifier_choices.append('Svm')
# except:
#     pass
#
# try:
#     from nltk.classify import scikitlearn
#     from sklearn.feature_extraction.text import TfidfTransformer
#     from sklearn.pipeline import Pipeline
#     from sklearn import ensemble, feature_selection, linear_model, naive_bayes, neighbors, svm, tree
#
#     classifiers = [
#         ensemble.ExtraTreesClassifier,
#         ensemble.GradientBoostingClassifier,
#         ensemble.RandomForestClassifier,
#         linear_model.LogisticRegression,
#         #linear_model.SGDClassifier, # NOTE: this seems terrible, but could just be the options
#         naive_bayes.BernoulliNB,
#         naive_bayes.GaussianNB,
#         naive_bayes.MultinomialNB,
#         neighbors.KNeighborsClassifier,  # TODO: options for nearest neighbors
#         svm.LinearSVC,
#         svm.NuSVC,
#         svm.SVC,
#         tree.DecisionTreeClassifier,
#     ]
#     sklearn_classifiers = {}
#
#     for classifier in classifiers:
#         sklearn_classifiers[classifier.__name__] = classifier
#
#     classifier_choices.extend(sorted(['sklearn.%s' % c.__name__ for c in classifiers]))
# except ImportError as exc:
#     sklearn_classifiers = {}

def train_classifier(input_dict):
    classifier=input_dict['classifier']
    training_bow_dataset = input_dict['training_data'] #BowDataset
    training_data=training_bow_dataset.bow_in_proper_format(classifier)

    if isinstance(classifier,NltkClassifier):
        trained_classifier=classifier.train(training_data)
        return {'trained_classifier': trained_classifier}
    elif input_dict['classifier'].__module__.startswith('sklearn'):
        classifier.fit(training_data, training_bow_dataset.labels)
        return {'trained_classifier': classifier}
    else:
        from tf_latino.latino.library_gen import latino_train_classifier
        return latino_train_classifier(input_dict)

def convert_to_probdists(csf,y_probas):
    classes = csf.classes_
    return [DictionaryProbDist(dict((classes[i], p)
                                   for i, p in enumerate(y_proba))) for y_proba in y_probas]

#apply_classifier already exists in orange package
def apply_bow_classifier(input_dict):
    trained_classifier = input_dict['trained_classifier']
    if input_dict.get('probability','true')=='true':
        try:
            trained_classifier.set_params(probability=True)
        except (ValueError,AttributeError):   #some classifiers don't have probability parameter
            pass

    testing_bow_dataset = input_dict['testing_dataset']
    testing_dataset=testing_bow_dataset.bow_in_proper_format(trained_classifier,no_labels=True)

    classifier_package=input_dict['trained_classifier'].__module__
    if trained_classifier.__class__.__name__=='LatinoObject':  #check if this is a latino object
        from tf_latino.latino.library_gen import latino_predict_classification
        return latino_predict_classification(input_dict)
    elif classifier_package.startswith("sklearn"):
        #a=trained_classifier.predict(testing_dataset)
        #example: http://scikit-learn.org/stable/auto_examples/document_classification_20newsgroups.html
        try:
            results = [DictionaryProbDist.from_probabilities_and_classes(example_predictions,trained_classifier.classes_)
                       for example_predictions in trained_classifier.predict_proba(testing_dataset)]
        except AttributeError:
            results = [DictionaryProbDist.from_prediction_and_classes(example_prediction,trained_classifier.classes_)
                       for example_prediction in trained_classifier.predict(testing_dataset)]
        #results=convert_to_probdists(trained_classifier,_results)
    elif isinstance(trained_classifier,NltkClassifier):
        results=[DictionaryProbDist(prob_dict=dpd._prob_dict,normalize=True)
                 for dpd in trained_classifier.prob_classify_many(testing_dataset)]
    else:
        raise Exception("What are you connecting me to then?")

    return {'labeled_dataset': None, 'predictions': results}



def extract_classifier_name(input_dict):
    import re

    in1=input_dict['classifier']
    clsf=in1.__class__.__module__+'.'+in1.__class__.__name__ if not in1.__class__.__name__=='LatinoObject' else in1.name


    out2=re.search(r'[A-Za-z\.0-9]+',clsf).group()

    spl=out2.split('.')

    if spl[0]=='sklearn':
      spl[0]='scikit-learn'
    elif spl[0]=='Latino':
      spl[0]='LATINO'
    if spl[-1]=='naive':
      spl[-1]='Gaussian Naive Bayes Classifier'
    if spl[-1]=='LinearSVC':
      spl[-1]='SVM Linear Classifier'

    return {'classifier_name': '['+spl[0]+'] '+spl[-1]}


def extract_actual_and_predicted_values(input_dict):
    actual=input_dict['dataset'].labels
    predicted=[a.max() for a in input_dict['predictions']]
    return {'actual_and_predicted': [actual, predicted]}


import pickle

from tf_core.document_corpus import DocumentCorpus
from tagging_common import universal_sentence_tagger_hub
#from tagging_common_parallel import universal_sentence_tagger_hub
#from nltk.tag.simplify   import (simplify_brown_tag, simplify_wsj_tag,
#                                 simplify_indian_tag, simplify_alpino_tag,
#                                 simplify_tag)
from nltk.tag.sequential import (DefaultTagger, NgramTagger, AffixTagger,
                                 RegexpTagger, #<--TODO
                                 ClassifierBasedPOSTagger)
import nltk.tag.brill
from nltk.tag.brill      import BrillTagger
from nltk.tag.brill_trainer import BrillTaggerTrainer
from nltk.tag.tnt        import TnT
from nltk.tag.hunpos     import HunposTagger
#from nltk.tag.crf        import MalletCRF
from django.conf import settings
from nltk.corpus import brown, treebank, nps_chat
from nltk.tag.stanford import StanfordPOSTagger
import os
import re


def pos_tagger_hub(input_dict):
    if input_dict['pos_tagger'].__class__.__name__=="LatinoObject": #check if this is a latino object
        from tf_latino.latino.library_gen import latino_pos_tag
        from workflows.tasks import executeFunction

        adc= executeFunction.apply_async([latino_pos_tag,input_dict],queue="windows").wait()['adc'] \
            if settings.USE_WINDOWS_QUEUE else latino_pos_tag(input_dict)
    else:
        adc= universal_sentence_tagger_hub(input_dict)['adc']

    number_of_letters=int(input_dict['num_of_letters'])
    if number_of_letters!=-1:
        element_annotation_name = input_dict['element_annotation']
        output_annotation_name = input_dict['output_feature']
        for doc in adc.documents:
            for annotation in doc.get_annotations(element_annotation_name):
                if not output_annotation_name in annotation.features:
                    print input_dict['pos_tagger'],annotation.features
                    print doc.features
                else:
                    annotation.features[output_annotation_name]=annotation.features[output_annotation_name][0:number_of_letters]

    return {'adc': adc }


def extract_pos_tagger_name(input_dict):
    tagger=input_dict['pos_tagger']
    if isinstance(tagger, dict) and 'name' in tagger:
        tagger_name=tagger['name']
    else:
        tagger_name=tagger['object'].__class__.__name__ if not tagger.__class__.__name__=="LatinoObject" else tagger.name
        tagger_name=re.search(r'[A-Za-z\.0-9]+',tagger_name).group() #extracts valid characters
    return {'pos_tagger_name': tagger_name}


def corpus_reader(corpus, chunk):
    if type(corpus)==DocumentCorpus:
        raise NotImplementedError
    else:
        tagged_posts = getattr(corpus, "tagged_posts", None)
        reverse = False
        if chunk[0] == '^':
            reverse = True
            chunk = chunk[1:]
        if callable(tagged_posts):
            if chunk[-1] == '%':
                index = (float(chunk[:-1])/100) * len(corpus.tagged_posts())
            else:
                index = int(chunk)
            if reverse:
                return corpus.tagged_posts()[int(index):]
            return corpus.tagged_posts()[:int(index)]
        else:
            if chunk[-1] == '%':
                index = (float(chunk[:-1])/100) * len(corpus.tagged_sents())
            else:
                index = int(chunk)
            if reverse:
                return corpus.tagged_sents()[int(index):]
            return corpus.tagged_sents()[:int(index)]


def nltk_default_pos_tagger(input_dict):
    """
    A tagger that assigns the same tag to every token.

        >>> from nltk.tag.sequential import DefaultTagger
        >>> default_tagger = DefaultTagger('NN')
        >>> default_tagger.tag('This is a test'.split())
        [('This', 'NN'), ('is', 'NN'), ('a', 'NN'), ('test', 'NN')]

    This tagger is recommended as a backoff tagger, in cases where
    a more powerful tagger is unable to assign a tag to the word
    (e.g. because the word was not seen during training).

    :param default_tag: The default tag "%(default)s". Set this to a different tag, such as "NN",
            to change the default tag.

    :returns pos_tagger: A python dictionary containing the POS tagger
        object and its arguments.
    """
    from nltk.tag import DefaultTagger
    return {'pos_tagger': {
        'function':'tag_sents',
        'object': DefaultTagger(input_dict.get('default_tag','-None-'))
        }
    }

#SEQUENTIAL TAGGERS
def nltk_affix_pos_tagger(input_dict):
    """
    A tagger that chooses a token's tag based on a leading or trailing
    substring of its word string.  (It is important to note that these
    substrings are not necessarily "true" morphological affixes).  In
    particular, a fixed-length substring of the word is looked up in a
    table, and the corresponding tag is returned.  Affix taggers are
    typically constructed by training them on a tagged corpus.

    :param training_corpus: A tagged corpus included with NLTK, such as treebank, brown, cess_esp, floresta,
        or an Annotated Document Corpus in the standard TextFlows' adc format
    :param backoff_tagger: A backoff tagger, to be used by the new
        tagger if it encounters an unknown context.
    :param affix_length: The length of the affixes that should be
        considered during training and tagging.  Use negative
        numbers for suffixes.
    :param min_stem_length: Any words whose length is less than
        min_stem_length+abs(affix_length) will be assigned a
        tag of None by this tagger.

    :returns pos_tagger: A python dictionary containing the POS tagger object and its arguments.
    """
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    training_corpus=corpus_reader(corpus, chunk)
    backoff_tagger=input_dict['backoff_tagger']['object'] if input_dict['backoff_tagger'] else DefaultTagger('-None-')
    affix_length=int(input_dict['affix_length'])
    min_stem_length=int(input_dict['min_stem_length'])
    cutoff=int(input_dict['cutoff']) #default 0


    return {'pos_tagger': {
                'function':'tag_sents',
                'object': AffixTagger(training_corpus, affix_length=affix_length, cutoff=cutoff,
                         min_stem_length=min(min_stem_length, 2), backoff=backoff_tagger)
        }
    }



def nltk_ngram_pos_tagger(input_dict):
    """
    A tagger that chooses a token's tag based on its word string and
    on the preceding n word's tags.  In particular, a tuple
    (tags[i-n:i-1], words[i]) is looked up in a table, and the
    corresponding tag is returned.  N-gram taggers are typically
    trained on a tagged corpus.

    Train a new NgramTagger using the given training data or
    the supplied model.  In particular, construct a new tagger
    whose table maps from each context (tag[i-n:i-1], word[i])
    to the most frequent tag for that context.  But exclude any
    contexts that are already tagged perfectly by the backoff
    tagger.

    :param training_corpus: A tagged corpus included with NLTK, such as treebank, brown, cess_esp, floresta,
        or an Annotated Document Corpus in the standard TextFlows' adc format
    :param backoff_tagger: A backoff tagger, to be used by the new
        tagger if it encounters an unknown context.
    :param cutoff: If the most likely tag for a context occurs
        fewer than *cutoff* times, then exclude it from the
        context-to-tag table for the new tagger.
    :param n:  N-gram is a contiguous sequence of n items from a given sequence of text or speech.

    :returns pos_tagger: A python dictionary containing the POS tagger object and its arguments.
    """
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    training_corpus=corpus_reader(corpus, chunk)
    backoff_tagger=input_dict['backoff_tagger']['object'] if input_dict['backoff_tagger'] else DefaultTagger('-None-')
    n=int(input_dict['n']) #default 2
    cutoff=int(input_dict['cutoff']) #default 0

    return {'pos_tagger': {
                'function':'tag_sents',
                'object': NgramTagger(n, train=training_corpus, model=None,
                 backoff=backoff_tagger, cutoff=0)
            }
    }


# CLASSIFIER TAGGERS
def nltk_phonetic_classifier_based_pos_tagger(input_dict):
    raise NotImplementedError

def nltk_classifier_based_pos_tagger(input_dict):
    """
    A sequential tagger that uses a classifier to choose the tag for
    each token in a sentence.  The featureset input for the classifier
    is generated by a feature detector function::

        feature_detector(tokens, index, history) -> featureset

    Where tokens is the list of unlabeled tokens in the sentence;
    index is the index of the token for which feature detection
    should be performed; and history is list of the tags for all
    tokens before index.

    Construct a new classifier-based sequential tagger.

    :param training_corpus: A tagged corpus consisting of a list of tagged
        sentences, where each sentence is a list of (word, tag) tuples.
    :param backoff_tagger: A backoff tagger, to be used by the new tagger
        if it encounters an unknown context.

TODO: odloci se katerega se obdrzi od naslednjih dveh

    :param classifier_builder: A function used to train a new
        classifier based on the data in *train*.  It should take
        one argument, a list of labeled featuresets (i.e.,
        (featureset, label) tuples).
    :param classifier: The classifier that should be used by the
        tagger.  #This is only useful if you want to manually
        construct the classifier; normally, you would use *train*
        instead.
    :param backoff_tagger: A backoff tagger, used if this tagger is
        unable to determine a tag for a given token.
    :param cutoff_prob: If specified, then this tagger will fall
        back on its backoff tagger if the probability of the most
        likely tag is less than *cutoff_prob*.

    :returns pos_tagger: A python dictionary containing the POS tagger
        object and its arguments.
    """
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    training_corpus=corpus_reader(corpus, chunk)
    backoff_tagger=input_dict['backoff_tagger']['object'] if input_dict['backoff_tagger'] else DefaultTagger('-None-')
    classifier=None #(input_dict['classifier'])
    cutoff_prob=int(input_dict['cutoff_prob']) if input_dict['cutoff_prob'] else None

    import nltk
    tagger_object=ClassifierBasedPOSTagger(train=training_corpus, classifier=classifier,
                 backoff=backoff_tagger, cutoff_prob=cutoff_prob)
    return {'pos_tagger': {
                'function':'tag_sents',
                'object': tagger_object
            }
    }

#BRILL TAGGERS
def nltk_brill_pos_tagger(input_dict):
    """Brill's transformational rule-based tagger.  Brill taggers use an
    initial tagger (such as ``tag.DefaultTagger``) to assign an initial
    tag sequence to a text; and then apply an ordered list of
    transformational rules to correct the tags of individual tokens.
    These transformation rules are specified by the ``BrillRule``
    interface.

    Brill taggers can be created directly, from an initial tagger and
    a list of transformational rules; but more often, Brill taggers
    are created by learning rules from a training corpus, using either
    ``BrillTaggerTrainer`` or ``FastBrillTaggerTrainer``.

    :param training_corpus: A tagged corpus consisting of a list of tagged
        sentences, where each sentence is a list of (word, tag) tuples.
    :param initial_tagger: The initial tagger. Brill taggers use an initial
        tagger (such as ``DefaultTagger``) to assign an initial tag
        sequence to a text.
    :param max_rules: The maximum number of transformations to be created
    :param min_score: The minimum acceptable net error reduction
        that each transformation must produce in the corpus.
    :param deterministic: If true, then choose between rules that
        have the same score by picking the one whose __repr__
        is lexicographically smaller.  If false, then just pick the
        first rule we find with a given score -- this will depend
        on the order in which keys are returned from dictionaries,
        and so may not be the same from one run to the next.  If
        not specified, treat as true iff trace > 0.
    :param templates: templates to be used in training

    :returns pos_tagger: A python dictionary containing the POS tagger
        object and its arguments.
    """
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    training_corpus = corpus_reader(corpus, chunk)
    initial_tagger = input_dict['initial_tagger']['object'] if input_dict['initial_tagger'] else DefaultTagger('-None-')
    max_rules = int(input_dict['max_rules']) #default 200
    min_score = int(input_dict['min_score']) #default 2
    deterministic = True

    templates = getattr(nltk.tag.brill,input_dict['templates'])()

    trainer = BrillTaggerTrainer(initial_tagger, templates, deterministic=deterministic, trace=settings.DEBUG)
    brill_tagger = trainer.train(training_corpus, max_rules=max_rules, min_score=min_score) #return BrillTagger(self._initial_tagger, rules)

    if settings.DEBUG:
        for rule in brill_tagger.rules():
            print(str(rule))

    return {'pos_tagger': {
                'function':'tag_sents',
                'object': brill_tagger
            }
    }

import time, os
import re
from collections import defaultdict

from nltk import TaggerI, FreqDist, untag, config_megam
from nltk.classify.maxent import MaxentClassifier
                  

class MaxentPosTagger(TaggerI):

    """
    MaxentPosTagger is a part-of-speech tagger based on Maximum Entropy models.
    """
    def train(self, train_sents, algorithm='megam', rare_word_cutoff=5,
              rare_feat_cutoff=5, uppercase_letters='[A-Z]', trace=3,
              **cutoffs):
        self.uppercase_letters = uppercase_letters
        self.word_freqdist = self.gen_word_freqs(train_sents)
        self.featuresets = self.gen_featsets(train_sents,
                rare_word_cutoff)
        self.features_freqdist = self.gen_feat_freqs(self.featuresets)
        self.cutoff_rare_feats(self.featuresets, rare_feat_cutoff)

        t1 = time.time()
        self.classifier = MaxentClassifier.train(self.featuresets, algorithm,
                                                 trace, **cutoffs)
        t2 = time.time()
        if trace > 0:
            print "time to train the classifier: {0}".format(round(t2-t1, 3))

    def gen_feat_freqs(self, featuresets):
        features_freqdist = defaultdict(int)
        for (feat_dict, tag) in featuresets:
            for (feature, value) in feat_dict.items():
                features_freqdist[ ((feature, value), tag) ] += 1
        return features_freqdist

    def gen_word_freqs(self, train_sents):
        word_freqdist = FreqDist()
        for tagged_sent in train_sents:
            for (word, _tag) in tagged_sent:
                word_freqdist[word] += 1
        return word_freqdist

    def gen_featsets(self, train_sents, rare_word_cutoff):
        featuresets = []
        for tagged_sent in train_sents:
            history = []
            untagged_sent = untag(tagged_sent)
            for (i, (_word, tag)) in enumerate(tagged_sent):
                featuresets.append( (self.extract_feats(untagged_sent, i,
                    history, rare_word_cutoff), tag) )
                history.append(tag)
        return featuresets


    def cutoff_rare_feats(self, featuresets, rare_feat_cutoff):
        never_cutoff_features = set(['w','t'])

        for (feat_dict, tag) in featuresets:
            for (feature, value) in feat_dict.items():
                feat_value_tag = ((feature, value), tag)
                if self.features_freqdist[feat_value_tag] < rare_feat_cutoff:
                    if feature not in never_cutoff_features:
                        feat_dict.pop(feature)


    def extract_feats(self, sentence, i, history, rare_word_cutoff=5):
        features = {}
        hyphen = re.compile("-")
        number = re.compile("\d")
        uppercase = re.compile(self.uppercase_letters)

        #get features: w-1, w-2, t-1, t-2.
        #takes care of the beginning of a sentence
        if i == 0: #first word of sentence
            features.update({"w-1": "<START>", "t-1": "<START>",
                             "w-2": "<START>", "t-2 t-1": "<START> <START>"})
        elif i == 1: #second word of sentence
            features.update({"w-1": sentence[i-1], "t-1": history[i-1],
                             "w-2": "<START>",
                             "t-2 t-1": "<START> %s" % (history[i-1])})
        else:
            features.update({"w-1": sentence[i-1], "t-1": history[i-1],
                "w-2": sentence[i-2],
                "t-2 t-1": "%s %s" % (history[i-2], history[i-1])})

        #get features: w+1, w+2. takes care of the end of a sentence.
        for inc in [1, 2]:
            try:
                features["w+%i" % (inc)] = sentence[i+inc]
            except IndexError:
                features["w+%i" % (inc)] = "<END>"

        if self.word_freqdist[sentence[i]] >= rare_word_cutoff:
            #additional features for 'non-rare' words
            features["w"] = sentence[i]

        else: #additional features for 'rare' or 'unseen' words
            features.update({"suffix(1)": sentence[i][-1:],
                "suffix(2)": sentence[i][-2:], "suffix(3)": sentence[i][-3:],
                "suffix(4)": sentence[i][-4:], "prefix(1)": sentence[i][:1],
                "prefix(2)": sentence[i][:2], "prefix(3)": sentence[i][:3],
                "prefix(4)": sentence[i][:4]})
            if hyphen.search(sentence[i]) != None:
                #set True, if regex is found at least once
                features["contains-hyphen"] = True
            if number.search(sentence[i]) != None:
                features["contains-number"] = True
            if uppercase.search(sentence[i]) != None:
                features["contains-uppercase"] = True

        return features


    def tag(self, sentence, rare_word_cutoff=5):
        history = []
        for i in xrange(len(sentence)):
            featureset = self.extract_feats(sentence, i, history,
                                               rare_word_cutoff)
            tag = self.classifier.classify(featureset)
            history.append(tag)
        return zip(sentence, history)

    def tag_sents(self, sentences):
        return [self.tag(sent) for sent in sentences]


def nltk_maxent_pos_tagger(input_dict):
    name = 'MaxentPosTagger'
    if not input_dict['training_corpus']:
        maxent_tagger = nltk.data.load('taggers/maxent_treebank_pos_tagger/english.pickle')
        name += '-pretrained'
    else:
        nltk.config_megam(settings.MEGAM_EXECUTABLE_PATH)

        maxent_tagger = MaxentPosTagger()
        chunk = input_dict['training_corpus']['chunk']
        corpus = input_dict['training_corpus']['corpus']
        training_corpus=corpus_reader(corpus, chunk)
        if training_corpus:
            maxent_tagger.train(training_corpus)
        else:
            raise AttributeError

    return {'pos_tagger': {
                'function':'tag_sents',
                'object': maxent_tagger,
                'name': name
            }
    }

from nltk.tag.perceptron import PerceptronTagger

def nltk_perceptron_pos_tagger(input_dict):
    name= 'PerceptronPosTagger'
    if not input_dict['training_corpus']:
        perceptron_tagger = PerceptronTagger()
        name += '-pretrained'
    else: 
        perceptron_tagger = PerceptronTagger(load=False)
        chunk = input_dict['training_corpus']['chunk']
        corpus = input_dict['training_corpus']['corpus']
        training_corpus=corpus_reader(corpus, chunk)
        perceptron_tagger.train(list(training_corpus))

    return {'pos_tagger': {
                'function':'tag_sents',
                'object': perceptron_tagger,
                'name': name
            }
    }

import random
from collections import defaultdict
from nltk.data import find, load


class AveragedPerceptron(object):

    '''An averaged perceptron, as implemented by Matthew Honnibal.
    See more implementation details here:
        http://spacy.io/blog/part-of-speech-POS-tagger-in-python/
    '''

    def __init__(self):
        # Each feature gets its own weight vector, so weights is a dict-of-dicts
        self.weights = {}
        self.classes = set()
        # The accumulated values, for the averaging. These will be keyed by
        # feature/clas tuples
        self._totals = defaultdict(int)
        # The last time the feature was changed, for the averaging. Also
        # keyed by feature/clas tuples
        # (tstamps is short for timestamps)
        self._tstamps = defaultdict(int)
        # Number of instances seen
        self.i = 0


    def predict(self, features):
        '''Dot-product the features and current weights and return the best label.'''
        scores = defaultdict(float)
        for feat, value in features.items():
            if feat not in self.weights or value == 0:
                continue
            weights = self.weights[feat]
            for label, weight in weights.items():
                scores[label] += value * weight
        # Do a secondary alphabetic sort, for stability
        return max(self.classes, key=lambda label: (scores[label], label))


    def update(self, truth, guess, features):
        '''Update the feature weights.'''
        def upd_feat(c, f, w, v):
            param = (f, c)
            self._totals[param] += (self.i - self._tstamps[param]) * w
            self._tstamps[param] = self.i
            self.weights[f][c] = w + v

        self.i += 1
        if truth == guess:
            return None
        for f in features:
            weights = self.weights.setdefault(f, {})
            upd_feat(truth, f, weights.get(truth, 0.0), 1.0)
            upd_feat(guess, f, weights.get(guess, 0.0), -1.0)


    def average_weights(self):
        '''Average weights from all iterations.'''
        for feat, weights in self.weights.items():
            new_feat_weights = {}
            for clas, weight in weights.items():
                param = (feat, clas)
                total = self._totals[param]
                total += (self.i - self._tstamps[param]) * weight
                averaged = round(total / self.i, 3)
                if averaged:
                    new_feat_weights[clas] = averaged
            self.weights[feat] = new_feat_weights


    def save(self, path):
        '''Save the pickled model weights.'''
        with open(path, 'wb') as fout:
            return pickle.dump(dict(self.weights), fout)


    def load(self, path):
        '''Load the pickled model weights.'''
        self.weights = load(path)



class CustomPerceptronTagger(TaggerI):

    '''
    Greedy Averaged Perceptron tagger, as implemented by Matthew Honnibal with a fix that
    makes it not crash on 0 lenth tokens during training.
    See more implementation details here:
        http://spacy.io/blog/part-of-speech-POS-tagger-in-python/
    
    >>> from nltk.tag.perceptron import PerceptronTagger
    Train the model 
    
    >>> tagger = PerceptronTagger(load=False)
    
    >>> tagger.train([[('today','NN'),('is','VBZ'),('good','JJ'),('day','NN')],
    ... [('yes','NNS'),('it','PRP'),('beautiful','JJ')]])
    
    >>> tagger.tag(['today','is','a','beautiful','day'])
    [('today', 'NN'), ('is', 'PRP'), ('a', 'PRP'), ('beautiful', 'JJ'), ('day', 'NN')]
    
    Use the pretrain model (the default constructor) 
    
    >>> pretrain = PerceptronTagger()
    
    >>> pretrain.tag('The quick brown fox jumps over the lazy dog'.split())
    [('The', 'DT'), ('quick', 'JJ'), ('brown', 'NN'), ('fox', 'NN'), ('jumps', 'VBZ'), ('over', 'IN'), ('the', 'DT'), ('lazy', 'JJ'), ('dog', 'NN')]
    
    >>> pretrain.tag("The red cat".split())
    [('The', 'DT'), ('red', 'JJ'), ('cat', 'NN')]
    '''

    START = ['-START-', '-START2-']
    END = ['-END-', '-END2-']
    
    def __init__(self, load=True):
        '''
        :param load: Load the pickled model upon instantiation.
        '''
        self.model = AveragedPerceptron()
        self.tagdict = {}
        self.classes = set()
        if load:
            PICKLE = "averaged_perceptron_tagger.pickle"
            AP_MODEL_LOC = 'file:'+str(find('taggers/averaged_perceptron_tagger/'+PICKLE))
            self.load(AP_MODEL_LOC)


    def tag(self, tokens):
        '''
        Tag tokenized sentences.
        :params tokens: list of word
        :type tokens: list(str)
        '''
        prev, prev2 = self.START
        output = []
        
        context = self.START + [self.normalize(w) for w in tokens] + self.END
        for i, word in enumerate(tokens):
            tag = self.tagdict.get(word)
            if not tag:
                features = self._get_features(i, word, context, prev, prev2)
                tag = self.model.predict(features)
            output.append((word, tag))
            prev2 = prev
            prev = tag

        return output


    def _pc(n, d):
        return (n / d) * 100



    def train(self, sentences, save_loc=None, nr_iter=5):
        '''Train a model from sentences, and save it at ``save_loc``. ``nr_iter``
        controls the number of Perceptron training iterations.
        :param sentences: A list of (words, tags) tuples.
        :param save_loc: If not ``None``, saves a pickled model in this location.
        :param nr_iter: Number of training iterations.
        '''
        self._make_tagdict(sentences)
        self.model.classes = self.classes
        for iter_ in range(nr_iter):
            c = 0
            n = 0
            for sentence  in sentences:
                words = [word for word,tag in sentence]
                tags  = [tag for word,tag in sentence]
                
                prev, prev2 = self.START
                context = self.START + [self.normalize(w) for w in words] + self.END
                for i, word in enumerate(words):
                    guess = self.tagdict.get(word)
                    if not guess:
                        feats = self._get_features(i, word, context, prev, prev2)
                        guess = self.model.predict(feats)
                        self.model.update(tags[i], guess, feats)
                    prev2 = prev
                    prev = guess
                    c += guess == tags[i]
                    n += 1
            random.shuffle(list(sentences))
        self.model.average_weights()
        # Pickle as a binary file
        if save_loc is not None:
            with open(save_loc, 'wb') as fout:
                pickle.dump((self.model.weights, self.tagdict, self.classes), fout, -1)
        

    def load(self, loc):
        '''
        :param loc: Load a pickled model at location.
        :type loc: str 
        '''

        self.model.weights, self.tagdict, self.classes = load(loc)
        self.model.classes = self.classes
        

    def normalize(self, word):
        '''
        Normalization used in pre-processing.
        - All words are lower cased
        - Digits in the range 1800-2100 are represented as !YEAR;
        - Other digits are represented as !DIGITS
        :rtype: str
        '''
        if len(word) == 0:
            return ""
        elif '-' in word and word[0] != '-':
            return '!HYPHEN'
        elif word.isdigit() and len(word) == 4:
            return '!YEAR'
        elif word[0].isdigit():
            return '!DIGITS'
        else:
            return word.lower()

    def _get_features(self, i, word, context, prev, prev2):
        '''Map tokens into a feature representation, implemented as a
        {hashable: float} dict. If the features change, a new model must be
        trained.
        '''
        def add(name, *args):
            features[' '.join((name,) + tuple(args))] += 1

        i += len(self.START)
        features = defaultdict(int)
        # It's useful to have a constant feature, which acts sort of like a prior
        add('bias')
        add('i suffix', word[-3:])
        if len(word) > 0:
            add('i pref1', word[0])
        add('i-1 tag', prev)
        add('i-2 tag', prev2)
        add('i tag+i-2 tag', prev, prev2)
        add('i word', context[i])
        add('i-1 tag+i word', prev, context[i])
        add('i-1 word', context[i-1])
        add('i-1 suffix', context[i-1][-3:])
        add('i-2 word', context[i-2])
        add('i+1 word', context[i+1])
        add('i+1 suffix', context[i+1][-3:])
        add('i+2 word', context[i+2])
        return features

    def _make_tagdict(self, sentences):
        '''
        Make a tag dictionary for single-tag words.
        :param sentences: A list of list of (word, tag) tuples.
        '''
        counts = defaultdict(lambda: defaultdict(int))
        for sentence in sentences:
            for word, tag in sentence:
                counts[word][tag] += 1
                self.classes.add(tag)
        freq_thresh = 20
        ambiguity_thresh = 0.97
        for word, tag_freqs in counts.items():
            tag, mode = max(tag_freqs.items(), key=lambda item: item[1])
            n = sum(tag_freqs.values())
            # Don't add rare words to the tag dictionary
            # Only add quite unambiguous words
            if n >= freq_thresh and (mode / n) >= ambiguity_thresh:
                self.tagdict[word] = tag

def nltk_custom_perceptron_pos_tagger(input_dict):
    name= 'PerceptronPosTagger'
    if not input_dict['training_corpus']:
        perceptron_tagger = CustomPerceptronTagger()
        name+='-pretrained'
    else: 
        perceptron_tagger = CustomPerceptronTagger(load=False)
        chunk = input_dict['training_corpus']['chunk']
        corpus = input_dict['training_corpus']['corpus']
        training_corpus=corpus_reader(corpus, chunk)
        perceptron_tagger.train(list(training_corpus))

    return {'pos_tagger': {
                'function':'tag_sents',
                'object': perceptron_tagger,
                'name': name
            }
    }



import copy

from tf_core.nltoolkit.helpers import NltkCorpus
import time
import re


def nltk_corpus(input_dict):
    """Returns the nltk.corpus for the selected corpus name"""
    return {'corpus': {'corpus': NltkCorpus(input_dict['corpus_name']), 'chunk': input_dict.get('chunk','100%')}}


def extract_nltk_corpus_name(input_dict):
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    name = "NLTK corpus " + str(chunk)
    match = re.search(r"(\\\\|/)(\w+)'", str(corpus))
    if match:
        name = match.group(2) + " " + str(chunk)
    return {'name': name}


def statistics(input_dict):
    """
    Statistics of Annotated Document Corpus.

    :param adc: Annotated Document Corpus.
    :return doc_count (int): Number of Documents.
    :return feature_count (int): Number of Features.
    :return description (string): Statistics.
    """

    output_dict = {}
    output_dict["doc_count"] = len(input_dict["adc"].documents)
    output_dict["feature_count"] = len(input_dict["adc"].features)
    output_dict["description"] = "Number of documents is " + str(output_dict["doc_count"]) + \
                                 ", number of corpus features is " + str(output_dict["feature_count"]) + "."
    return output_dict


def extract_feature(input_dict):
    """
    Extract documents features.

    :param adc
    :param feature_name
    :return extracted features (list)
    """

    extracted_features = []
    for document in input_dict["adc"].documents:
        extracted_features.append(document.features.get(input_dict["feature_name"], None))

    return {"strings": extracted_features}


def add_feature(input_dict):
    """
    Add a feature to Annotated Document Corpus.

    :param adc: Annotated Document Corpus
    :param feature_name: the name of new feature
    :param feature_values: list of features values
    :param feature_value_prefix:  prefix to feature value

    :return: new adc
    """

    for i, document in enumerate(input_dict["adc"].documents):
        document.features[input_dict[u"feature_name"]] = input_dict[u"feature_value_prefix"] + str(input_dict[u"feature_values"][i])
    return {"adc":input_dict["adc"]}


def add_computed_document_features(input_dict):
    """
    TODO: Add a feature to Annotated Document Corpus.

    :param adc: Annotated Document Corpus
    :param feature_name: the name of new feature
    :param feature_computation: "New Feature Computatation
    :param feature_spec: Comma separated list of names of old features used in the 'New Feature Computataion'.

    :return: new adc
    """

    adc=input_dict["adc"]
    compute_new_features(adc.documents,input_dict["feature_name"],input_dict["feature_computation"])

    return {"adc":adc}

def add_computed_token_features(input_dict):
    """
    TODO

    :param adc: Annotated Document Corpus
    :param feature_name: the name of new feature
    :param feature_computation: "New Feature Computatation
    :param feature_spec: Comma separated list of names of old features used in the 'New Feature Computataion'.

    :return: new adc
    """

    adc=input_dict["adc"]
    for document in adc.documents:
        compute_new_features(document.get_annotations(input_dict["annotation_name"]),input_dict["feature_name"],input_dict["feature_computation"])

    return {"adc":adc}

def compute_new_features(objs,new_feature_name,feature_computation):
    features = re.findall(r"\{([\w\s]+)\}", feature_computation) #[mm.split(':')[0] for mm in m.groups()]
    for obj in objs:
        new_feature_value=copy.copy(feature_computation)
        for feature in features:
            new_feature_value=new_feature_value.replace("{"+feature+"}",obj.features.get(feature, "NULL"))
        obj.features[new_feature_name]=new_feature_value


def split_documents_by_feature_value(input_dict):
    """

    Split Annotated Document Corpus by conditions with features and values.

    :param adc: Annotated Document Corpus
    :param feature_condition: split condition.
    :param discard_filtered_out default-false: discard documents that do not fulfil given condition

    :return adc_filtered: adc with documents that fulfil the condition
    :return adc_rest: adc with documents that do not fulfil the condtion
    """
    discard_filtered_out = input_dict["discard_filtered_out"] == u"true"
    adc_filtered, adc_rest = None, input_dict["adc"]

    if input_dict["feature_condition"] != u'':
        feature_conditions = [el.strip().split("=") for el in re.split("[;|,]",input_dict["feature_condition"])]
        for i in range(len(feature_conditions)):
            if len(feature_conditions[i]) != 2:
                raise Exception("Feature condition should be specified like: feature1 = value1, feature2 = value2 etc.")
            feature, value = feature_conditions[i]
            feature_conditions[i] = [feature.strip(), value.strip()]

        included = []
        discarded = []
        for i, document in enumerate(input_dict["adc"].documents):
            if all([document.features.get(feature, "") == value for feature, value in feature_conditions]):
                included.append(i)
            else:
                discarded.append(i)

        adc_filtered, adc_rest = input_dict["adc"].split(included, None if discard_filtered_out else discarded)

    if discard_filtered_out:
        return {"adc_filtered":adc_filtered, "adc_rest": None}
    else:
        return {"adc_filtered": adc_filtered, "adc_rest": adc_rest}


def extract_documents(input_dict):
    """
    Extract documents, given document indices, from Annotated Document Corpus.

    :param adc: Annotated Document Corpus
    :param index_list: list of document ids to include
    :param discard_filtered_out default-false: discard documents that are not included in index_list

    :return adc_filtered: adc with documents that fulfil the condition
    :return adc_rest: adc with documents that do not fulfil the condtion
    """

    discard_filtered_out = input_dict["discard_filtered_out"] == u"true"
    included = input_dict["index_list"]

    for i in included:
        if i < 0 or i >= len(input_dict["adc"].documents) or type(i) != int:
            raise Exception("Document indices are set incorrectly.")
    discarded = [i for i in range(len(input_dict["adc"].documents)) if i not in input_dict["index_list"]]
    adc_filtered, adc_rest = input_dict["adc"].split(included, None if discard_filtered_out else discarded)

    return {"adc_filtered": adc_filtered, "adc_rest": adc_rest}


def merge_corpora(input_dict):
    """

    Merge multiple Annotated Document Corpuses into one.

    :param adc: list of Annotated Document Corpuses
    :return: Annotated Document Corpus
    """

    adc = input_dict["adc"][0]
    adc.features["SourceDate"] = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))
    adc.features["CorpusCreateDate"] = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))

    for i in range(1, len(input_dict["adc"])):
        for document in input_dict["adc"][i].documents:
            adc.documents.append(document)

    return {"adc": adc}

def extract_adc_name(input_dict):
    return {'name' : input_dict['adc'].features['Source']}





from django.conf import settings

from tagging_common import universal_word_tagger_hub

def stop_word_tagger_hub(input_dict):
    """
    Apply the *stop_word_tagger* object on the Annotated Document Corpus (*adc*):

    1. first select only annotations of type Token Annotation *element_annotation*,
    2. apply the stop_word tagger
    3. create new annotations *output_feature* with the outputs of the stop word tagger.

    :param adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    :param stop_word_tagger: A python dictionary containing the stop word tagger object and its arguments.
    :param element_annotation: Which annotated part of document to be searched for stopwords.
    :param output_features: How to annotate the newly discovered stop word features.

    :returns adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    """

    if input_dict['stop_word_tagger'].__class__.__name__=='LatinoObject':
        from workflows.tasks import executeFunction
        from tf_latino.latino.library_gen import latino_tag_adcstopwords
        input_dict['tagger']=input_dict['stop_word_tagger']  #TODO temporary
        return executeFunction.apply_async([latino_tag_adcstopwords,input_dict],queue="windows").wait() if settings.USE_WINDOWS_QUEUE \
            else latino_tag_adcstopwords(input_dict)

    else:
        adc = input_dict['adc']
        tagger_dict = input_dict['stop_word_tagger']
        input_annotation = input_dict['element_annotation']
        output_annotation = input_dict['output_feature']
        return universal_word_tagger_hub(adc,tagger_dict,input_annotation,output_annotation)



#STOPWORD TAGGERS
class StopWordTagger:
    def __init__(self,stop_words,ignore_case=True):
        self.ignore_case=ignore_case
        self.stop_words=stop_words
        if type(stop_words) in [str,unicode]:
            self.stop_words=self.stop_words.split("\n")
        if ignore_case:
            self.stop_words=[sw.lower() for sw in self.stop_words]

    def tag(self,token):
        return "true" if (token.lower() if self.ignore_case else token) in self.stop_words else None


def nltk_stop_word_tagger(input_dict):
    """
    Constructs a python stop word tagger object.

    :param stop_words: A list or string (stop words separated by new lines) of stop words.
    :param ignore_case: If true than words are marked as stop word regardless of their casing.

    :returns stop_word_tagger: A python dictionary containing the StopWordTagger object and its arguments.
    """

    return {'stop_word_tagger':
                {'object':StopWordTagger(input_dict.get('stop_words',''),input_dict.get('ignore_case',True)),
                 'function':'tag',
                }
    }
# import nltk
# # STOP_TYPES = ['DET', 'CNJ']
# text = "I have a python module installed on my system."
# # tokens = nltk.pos_tag(nltk.word_tokenize(text))
# # good_words = [w for w, wtype in token if wtype not in STOP_TYPES]
#
# nltk.pos_tag(nltk.word_tokenize(text))

# import nltk
#
# with open('sample.txt', 'r') as f:
#     sample = f.read()
#
# sentences = nltk.sent_tokenize(sample)
# tokenized_sentences = [nltk.word_tokenize(sentence) for sentence in sentences]
# tagged_sentences = [nltk.pos_tag(sentence) for sentence in tokenized_sentences]
# chunked_sentences = nltk.batch_ne_chunk(tagged_sentences, binary=True)
#
# def extract_entity_names(t):
#     entity_names = []
#
#     if hasattr(t, 'node') and t.node:
#         if t.node == 'NE':
#             entity_names.append(' '.join([child[0] for child in t]))
#         else:
#             for child in t:
#                 entity_names.extend(extract_entity_names(child))
#
#     return entity_names
#
# entity_names = []
# for tree in chunked_sentences:
#     # Print results per sentence
#     # print extract_entity_names(tree)
#
#     entity_names.extend(extract_entity_names(tree))
#
# # Print all entity names
# #print entity_names
#
# # Print unique entity names
# print set(entity_names)


import django.template.loader
from django.shortcuts import get_object_or_404
from django.contrib.auth.decorators import login_required
from workflows.models import Widget
import nltk
from nltk.collocations import *
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_selection import chi2

@login_required
def display_document_corpus(request, input_dict, output_dict, widget, narrow_doc='n'):
    """
    Display Document Corpus widget displays ADC (Annotated Document Corpus) structure.
    It shows a detail view for selected document with annotations.

    user runs a display corpus widget. We return index template, which calls a document_corpus function.
    """
    view = django.shortcuts.render(request, 'visualizations/document_corpus_index_page.html', {'widget': widget,
                                                                                               'input_dict': input_dict,
                                                                                               'narrow_doc': narrow_doc})
    return view

@login_required
def document_corpus(request, widget_id, document_id_from=0, document_id_to=-1,  narrow_doc='n'):
    """
    Function display a list of documents, if there is less than 100 documents.
    If there is more than 100 document, it groups documents in groups of 100 documents.
    """
    w = get_object_or_404(Widget, pk=widget_id)
    features = w.inputs.all()[0].value.features  # features like source_name, corpus_create_date etc.

    if document_id_to == -1:  # if there is default value for document_id_to
        documents = w.inputs.all()[0].value.documents  # get all documents
        back_url = ""
    else:
        #get interval of documents. document_id_from has -1, because we start counting documents from 1 and not 0.
        #document_id_to does not have -1, because we output documents to document x, with document x included.
        documents = w.inputs.all()[0].value.documents[int(document_id_from):int(document_id_to)]
        back_url = request.environ["HTTP_REFERER"]

    if len(documents) <= 100:
        #display 100 documents or less

        for i, document in enumerate(documents):
            document.additions = {"id": i+int(document_id_from)} #add document ids to the document object
            #count annotations for a document
            document.additions["basic_types"] = len(set([annotation.type for annotation in document.annotations]))


        view = django.shortcuts.render(request, 'visualizations/document_corpus.html', {'widget_id': widget_id,
                                                                                        'documents': documents,
                                                                                        "features": features,
                                                                                        'back_url': back_url,
                                                                                        'narrow_doc': narrow_doc})
    else:
        #group documents in groups of 100 documents
        document_catalog = []
        for i in range(len(documents)):
            if i % 100 == 0:
                interval = i+100 if i+100 < len(documents) else len(documents)
                last_index = interval-1 if i+100 < len(documents) else interval-1

                sum_features = sum([len(documents[j].features) for j in range(i, interval)])
                sum_annotations = sum([len(documents[j].annotations) for j in range(i, interval)])

                document_catalog.append({"first": (documents[i], i+1),  # first document in a group with id
                                         "last": (documents[last_index], interval),  # last document in a group with id
                                         "sum_annotations": sum_annotations,  # sum of annotations  for this group
                                         "sum_features": sum_features,  # sum of features  for this group
                                         "length": 100 if i+100 < len(documents) else len(documents)-i})

        view = django.shortcuts.render(request, 'visualizations/document_corpus_catalog.html', {'widget_id': widget_id,
                                                                                                'document_catalog': document_catalog,
                                                                                                "features": features,
                                                                                                'back_url': back_url,
                                                                                                'narrow_doc': narrow_doc})

    return view

@login_required
def document_page(request, widget_id, document_id, narrow_doc='n'):
    """
    Function displays details for a single document.
    """
    import random_colors
    w = get_object_or_404(Widget, pk=widget_id)
    document = w.inputs.all()[0].value.documents[int(document_id)]
    back_url = request.environ["HTTP_REFERER"]

    #create a new data structure for annotations that is more appropriate for django template language
    annotations = {}
    for annotation in document.annotations:

        annotations[annotation.type] = annotations.get(annotation.type, "") + \
                                       str(annotation.span_start) + "," + \
                                       str(annotation.span_end) + "," + \
            str(annotation.features)[1:-1].replace(",", "<br/>").replace(":", " =").replace("u'", "").replace("'", "").replace("u\"", "").replace("\"", "") \
            + ":"  # annotations for tipsy. We format string, so that tipsy shows it correctly

    #create a color for each annotation
    annotation_colors = random_colors.get_colors(len(annotations))
    for i, (k, v) in enumerate(annotations.iteritems()):
        annotations[k] = [v, annotation_colors[i]]

    view = django.shortcuts.render(request, 'visualizations/document_page.html', {'widget_id': widget_id,
                                                                                  'document': document,
                                                                                  'annotations': annotations,
                                                                                  'back_url': back_url,
                                                                                  'narrow_doc': narrow_doc})
    return view


@login_required
def display_annotation_statistic(request, input_dict, output_dict, widget, narrow_doc='n'):
    """
    Display POS statistics, basicaly frequencies of specific tags
    """

    adc = input_dict['adc']
    annotation_name = input_dict['annotation_name']
    stat_type = input_dict['stat_type']
    allAnnotations = 0
    result_list = []
    n = int(input_dict['n_gram'])
    if stat_type == 'frequency':
        annotation_dict = {}
        for doc in adc.documents:
            annotations = doc.get_annotations_with_text(annotation_name)
            length = len(annotations)
            for i in range(0, length - n + 1):
                combo = ""
                for j in range(i, i + n):
                    _, value = annotations[j]
                    if j > i:
                        combo += "_"
                    combo += value
                            
                if len(combo) > 0:
                    allAnnotations += 1
                    if combo in annotation_dict:
                        annotation_dict[combo] = annotation_dict[combo] + 1
                    else:
                        annotation_dict[combo] = 1
            title = "Annotation frequency"
        
        allAnnotations = float(allAnnotations)
        for pos, number in annotation_dict.items():
            result_list.append((pos, "{0:.2f}".format(float(number)/allAnnotations)))

        result_list = sorted(result_list, key=lambda x: x[1], reverse=True)[:40]

    else:
        all_annotations = []
        for doc in adc.documents:
            annotations = doc.get_annotations_with_text(annotation_name)
            for annotation in annotations:
                _, value = annotation
                all_annotations.append(value)

        if stat_type == 'pmi_bigrams':
            bigram_measures = nltk.collocations.BigramAssocMeasures()
            finder = BigramCollocationFinder.from_words(all_annotations)
            best = finder.score_ngrams(bigram_measures.pmi)[:40]
            for tags, score in best:
                tag1, tag2 = tags
                result_list.append((tag1 + " " + tag2, "{0:.2f}".format(score)))
            title = "Bigrams PMI"


        elif stat_type == 'pmi_trigrams':
            trigram_measures = nltk.collocations.TrigramAssocMeasures()
            finder = TrigramCollocationFinder.from_words(all_annotations)
            best = finder.score_ngrams(trigram_measures.pmi)[:40]
            for tags, score in best:
                tag1, tag2, tag3 = tags
                result_list.append((tag1 + " " + tag2 + " " + tag3, "{0:.2f}".format(score)))
            title = "Trigrams PMI"

        elif stat_type == "chi_square_test":
            vectorizer = CountVectorizer(lowercase=False,stop_words=None, ngram_range=(n,n))
            raw_text = [document.raw_text(selector=annotation_name,
                                  join_annotations_with=" ",
                                  stop_word_feature_name="StopWord")
                for document in adc.documents]

            labels = [document.get_first_label() for document in adc.documents]
            X = vectorizer.fit_transform(raw_text)
            chi2score = chi2(X,labels)
            chi2score = zip(chi2score[0], chi2score[1])
            wscores = zip(vectorizer.get_feature_names(),chi2score)
            result_list = sorted(wscores, reverse=True, key=lambda x:x[1][0])[:40]
            result_list = [(feat, '{:.2f}'.format(chi[0]), '{:.2e}'.format(chi[1]))for feat, chi in result_list]
            title = "Chi square feature evaluation"

    view = django.shortcuts.render(request, 'visualizations/annotation_statistics.html', {'widget': widget,
                                                                                               'data': [result_list, title],
                                                                                               'narrow_doc': narrow_doc})
    return view






import copy
import json

from document import Document


class DocumentCorpus:
    def __init__(self, documents,features):
        self.documents=documents
        self.features=features

    def __unicode__(self):
        #for i in itemDir:
        return 'Documents; {0}' % (self.documents)

    def split(self,train_indices,test_indices=None):
        output_train = DocumentCorpus(copy.deepcopy([self.documents[i] for i in train_indices]),
                                      copy.deepcopy(self.features))
        output_test = DocumentCorpus(copy.deepcopy([self.documents[i] for i in test_indices]),
                                      copy.deepcopy(self.features)) if test_indices else None

        return output_train,output_test

    def get_labels(self):
        return json.loads(self.features['Labels']) if 'Labels' in self.features else []

    def get_document_labels(self):
        return [doc.get_first_label() for doc in self.documents]

    def __getstate__(self):
        print "get state!!"
        minimized_docs=[d.__minimize__() for d in self.documents]
        return json.dumps([minimized_docs,self.features])

    def __setstate__(self,value):
        print "set state!!"
        minimized_docs,self.features=json.loads(value)
        self.documents=[Document.__from_minimized__(d) for d in minimized_docs]






def get_new_color(colors):
    """
    function takes list of colors and it calculates new color, which is similar to others.

    :param colors: list of colors that are used as basis for calculating similarity.
    :return: color in a tuple, that has 3 values for RGB. Example: (200, 100, 200)
    """
    import random
    import math
    max_distance = 0

    for t in range(100):
        min_distance = 1000
        # randomly creates new color (R, G, B)
        new_color = (random.randint(0, 255), random.randint(0, 255), random.randint(0, 255))

        for color in colors:  # for each color calculate distances
            distance = math.sqrt((color[0] - new_color[0])**2 +
                                 (color[1] - new_color[1])**2 +
                                 (color[2] - new_color[2])**2)

            if distance < min_distance:
                min_distance = distance

        if min_distance > max_distance:
            max_distance = min_distance
            similar_color = new_color  # new color is the one with the minimal distance from others
    return similar_color


def get_colors(n):
    """
    function randomly creates n colors that are similar
    :param n: number of colors
    :return: list of html color codes
    """

    colors = [(0, 0, 0), (255, 255, 255)]  # init black, white
    for i in range(n):
        colors.append(get_new_color(colors))

    #create html color codes without init colors
    return ["#%02x%02x%02x" % color for color in colors[2:]]



from django.conf.urls import url

from tf_core.nltoolkit.visualization_views import document_corpus, document_page

"""
module cotains url redirections

:param widget_id: unique identifier of a widget globally in textflows. It used to separate views from others.
:param narrow_doc: it used to select a css style template, if the view is displayed as frame or in a new tab.
:param document_id_from: id of first document. It is used, if there is more than 100 documents.
:param document_id_to: id of last document. It is used, if there is more than 100 documents.
:param document_id: id of a document. It is used to output details of a document.
"""

urlpatterns = [
    #display a first view of document corpus, when user runs a widget
    url(r'^get-adc-index/widget(?P<widget_id>[0-9]+)/nx/Index.html$', document_corpus, name='nltk get adc index'),
    #display document corpus, when user clicks on a link: Open view in new window
    url(r'^get-adc-index/widget(?P<widget_id>[0-9]+)/(?P<narrow_doc>n?)x/Index.html$', document_corpus, name='nltk get adc index'),
    #if there is more than 100 documents, we display a catalog view, which groups documents in groups with 100 documents.
    url(r'^get-adc-index/widget(?P<widget_id>[0-9]+)/(?P<narrow_doc>n?)x/Index(?P<document_id_from>[0-9]+)-(?P<document_id_to>[0-9]+).html$', document_corpus, name='nltk get adc index'),
    #displays a document_page.html with details for a single document.
    url(r'^get-adc-index/widget(?P<widget_id>[0-9]+)/(?P<narrow_doc>n?)x/Document(?P<document_id>[0-9]+).html', document_page, name='nltk get adc page')
]



import nltk

# with open('football.txt', 'r') as f:
#     sample = f.read().encode('ascii', 'ignore')
#
# sentences = nltk.sent_tokenize(sample)
sentences=["CNN.com delivers the latest breaking news and information on the latest top stories, weather, business, entertainment, politics, and more",
           "Flights to London's Gatwick Airport were diverted and departures delayed for a time Monday after a Virgin Atlantic plane made an emergency landing, the airport said."]
tokenized_sentences = [nltk.word_tokenize(sentence) for sentence in sentences]
tagged_sentences = [nltk.pos_tag(sentence) for sentence in tokenized_sentences]
chunked_sentences = nltk.ne_chunk_sents(tagged_sentences, binary=True)


def extract_entity_names(t):
    entity_names = []

    if hasattr(t, 'label') and t.label:
        if t.label() == 'NE':
            entity_names.append(' '.join([child[0] for child in t]))
        else:
            for child in t:
                entity_names.extend(extract_entity_names(child))

    return entity_names

entity_names = []
for tree in chunked_sentences:
    # Print results per sentence
    # print extract_entity_names(tree)

    entity_names.extend(extract_entity_names(tree))

# Print all entity names
#print entity_names

# Print unique entity names
print set(entity_names)



import re
import os.path
from docx import opendocx, getdocumenttext
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from pdfminer.pdfpage import PDFPage
from cStringIO import StringIO
from subprocess import Popen, PIPE


def convert_pdf_to_txt(path):
    rsrcmgr = PDFResourceManager()
    retstr = StringIO()
    codec = 'utf-8'
    laparams = LAParams()
    device = TextConverter(rsrcmgr, retstr, codec=codec, laparams=laparams)
    fp = file(path, 'rb')
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    password = ""
    maxpages = 0
    caching = True
    pagenos = set()

    for page in PDFPage.get_pages(fp, pagenos, maxpages=maxpages, password=password, caching=caching, check_extractable=True):
        interpreter.process_page(page)
    fp.close()
    device.close()
    str = retstr.getvalue()
    retstr.close()
    return str


def document_to_text(path):
    ext = os.path.splitext(path)[1]
    filename = os.path.basename(path)

    if ext == ".doc":
        cmd = os.path.dirname(os.path.realpath(__file__)) + os.sep + "antiword" + os.sep + "antiword.exe -m CP852 " + path
        pipe = Popen(cmd, stdout=PIPE, shell=True)
        text = pipe.communicate()[0]
        return filename, unicode(re.sub("\r|\n", "", text).strip())

    elif ext == ".docx":
        document = opendocx(path)
        paratextlist = getdocumenttext(document)
        newparatextlist = []
        for paratext in paratextlist:
            newparatextlist.append(paratext)
        return filename, "".join(newparatextlist)

    elif ext == ".pdf":
        text = convert_pdf_to_txt(path)
        return filename, unicode(re.sub("\r|\n", "", text).strip())

    else:
        text = unicode(open(path, "r").read(), errors='replace').strip()
        return filename, text




from django.conf import settings

import nltk
from tagging_common import universal_word_tagger_hub
from nltk.corpus import wordnet
from pattern.vector import stem, PORTER, LEMMA

#from tagging_common_parallel import universal_word_tagger_hub
from tf_core.nltoolkit.helpers import NltkRegexpStemmer


def stem_lemma_tagger_hub(input_dict):
    if input_dict['tagger'].__class__.__name__=="LatinoObject": #check if this is a latino object
        from tf_latino.latino.library_gen import latino_tag_adc_stem_lemma
        from workflows.tasks import executeFunction

        return latino_tag_adc_stem_lemma(input_dict) if not settings.USE_WINDOWS_QUEUE \
            else executeFunction.apply_async([latino_tag_adc_stem_lemma,input_dict],queue="windows").wait()
    else:
        adc = input_dict['adc']
        tagger_dict = input_dict['tagger']
        input_annotation = input_dict['element_annotation']
        pos_annotation = input_dict.get('pos_annotation')
        output_annotation = input_dict['output_feature']
        return universal_word_tagger_hub(adc,tagger_dict,input_annotation,output_annotation,pos_annotation)

# STEMMERS
def nltk_lancaster_stemmer(input_dict):
    """
    A word stemmer based on the Lancaster stemming algorithm.
    """
    return {'tagger':
                {'object': nltk.stem.lancaster.LancasterStemmer(),
                 'function': 'stem', }
    }


def nltk_porter_stemmer(input_dict):
    """
    A word stemmer based on the Porter stemming algorithm.
    """
    return {'tagger':
                {'object': nltk.stem.porter.PorterStemmer(),
                 'function': 'stem',
                }
    }


def nltk_isri_stemmer(input_dict):
    """
    ISRI Arabic stemmer based on algorithm: Arabic Stemming without a root dictionary.
    Information Science Research Institute. University of Nevada, Las Vegas, USA.
    """
    return {'tagger':
                {'object': nltk.stem.isri.ISRIStemmer(),
                 'function': 'stem',
                }
    }


def nltk_regexp_stemmer(input_dict):
    """
    A stemmer that uses regular expressions to identify morphological
    affixes.  Any substrings that match the regular expressions will
    be removed.

    :type regexp: str or regexp
    :param regexp: The regular expression that should be used to
        identify morphological affixes.
    :type min: int
    :param min: The minimum length of string to stem
    """

    regexp = input_dict["regexp"] #default ing$|s$|e$|able$
    min = int(input_dict["min"]) #default 4
    return {'tagger':
                {'object': NltkRegexpStemmer(regexp, min=min),
                 'function': 'stem',
                }
    }


def nltk_rslp_stemmer(input_dict):
    """
   A stemmer for Portuguese.

   """
    return {'tagger':
                {'object': nltk.stem.rslp.RSLPStemmer(),
                 'function': 'stem',
                }}


def nltk_snowball_stemmer(input_dict):
    """
    Snowball Stemmer
    The following languages are supported:
    Danish, Dutch, English, Finnish, French, German,
    Hungarian, Italian, Norwegian, Portuguese, Romanian, Russian,
    Spanish and Swedish.

    The algorithm for English is documented here:
    Porter, M. \"An algorithm for suffix stripping.\"
    Program 14.3 (1980): 130-137.

    The algorithms have been developed by Martin Porter.
    These stemmers are called Snowball, because Porter created
    a programming language with this name for creating
    new stemming algorithms. There is more information available
    at http://snowball.tartarus.org/

    :param language: The language whose subclass is instantiated.
    :type language: str or unicode
    :param ignore_stopwords: If set to True, stopwords are
                         not stemmed and returned unchanged.
                         Set to False by default.
    :type ignore_stopwords: bool
    """

    language = input_dict["language"]
    ignore_stopwords = input_dict["ignore_stopwords"] == "true"

    if language == "danish":
        stemmer = nltk.stem.snowball.DanishStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "dutch":
        stemmer = nltk.stem.snowball.DutchStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "english":
        stemmer = nltk.stem.snowball.EnglishStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "finnish":
        stemmer = nltk.stem.snowball.FinnishStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "french":
        stemmer = nltk.stem.snowball.FrenchStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "german":
        stemmer = nltk.stem.snowball.GermanStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "hungarian":
        stemmer = nltk.stem.snowball.HungarianStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "italian":
        stemmer = nltk.stem.snowball.ItalianStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "norwegian":
        stemmer = nltk.stem.snowball.NorwegianStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "portuguese":
        stemmer = nltk.stem.snowball.PortugueseStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "romanian":
        stemmer = nltk.stem.snowball.RomanianStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "russian":
        stemmer = nltk.stem.snowball.RussianStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "spanish":
        stemmer = nltk.stem.snowball.SpanishStemmer(ignore_stopwords=ignore_stopwords)
    elif language == "swedish":
        stemmer = nltk.stem.snowball.SwedishStemmer(ignore_stopwords=ignore_stopwords)

    return {'tagger':
            {'object': stemmer,
             'function': 'stem',
            }}


class WordnetLemmatizer:
    def __init__(self, pos_annotation):
        self.pos_annotation = pos_annotation
        self.lemmatizer = nltk.stem.wordnet.WordNetLemmatizer()
        self.morphy_tag = {'NN':wordnet.NOUN, 'NNS':wordnet.NOUN,
                  'NNP':wordnet.NOUN, 'NNPS':wordnet.NOUN, 'JJ':wordnet.ADJ,
                  'JJR':wordnet.ADJ, 'JJS':wordnet.ADJ, 'VB':wordnet.VERB,
                  'VBD':wordnet.VERB, 'VBG':wordnet.VERB, 'VBN':wordnet.VERB,
                  'VBP':wordnet.VERB, 'VBZ':wordnet.VERB,'RB':wordnet.ADV,
                  'RBR':wordnet.ADV, 'RBS':wordnet.ADV}
    
    def lemmatize(self, lemma, **kwargs):
        if kwargs and self.pos_annotation:
            pos_tag = kwargs[self.pos_annotation]
            if pos_tag in self.morphy_tag:
                return self.lemmatizer.lemmatize(lemma, self.morphy_tag[pos_tag])
        return self.lemmatizer.lemmatize(lemma)



def nltk_wordnet_lemmatizer(input_dict):
    """
    WordNet Lemmatizer
    Lemmatize using WordNet's built-in morphy function.
    Returns the input word unchanged if it cannot be found in WordNet.
    """

    pos_annotation = input_dict.get('pos_annotation')
    return {'tagger':
                {'object': WordnetLemmatizer(pos_annotation),
                 'function': 'lemmatize',
                }}


class PatternLemmatizer:
    def lemmatize(self, word):
        return stem(word, stemmer = LEMMA)


def pattern_lemmatizer(input_dict):
    """
    WordNet Lemmatizer
    Lemmatize using WordNet's built-in morphy function.
    Returns the input word unchanged if it cannot be found in WordNet.
    """
    return {'tagger':
                {'object': PatternLemmatizer(),
                 'function': 'lemmatize',
                }}


class PatternPorterStemmer:
    def stem(self, word):
        return stem(word, stemmer = PORTER)


def pattern_porter_stemmer(input_dict):
    """
    WordNet Lemmatizer
    Lemmatize using WordNet's built-in morphy function.
    Returns the input word unchanged if it cannot be found in WordNet.
    """
    return {'tagger':
                {'object': PatternPorterStemmer(),
                 'function': 'stem',
                }}


class DefaultLemmatizer:
    #Baseline lemmatizer, doesn't do anything, returns unchanged word
    def lemmatize(self, word):
        return word


def default_lemmatizer(input_dict):
    return {'tagger':
                {'object': DefaultLemmatizer(),
                 'function': 'lemmatize',
                }}


































import json
from collections import defaultdict

from tf_core.annotation import Annotation


class Document:
    def __init__(self, name,text,annotations,features):
        self.annotations=annotations
        self.features=features
        self.name=name

        self.text=text

    def __unicode__(self):
        return 'Name; {0}\nText: {1}' % (self.name, self.text)

    def __minimize__(self):
        state=self.__dict__.copy()
        annotations=state.pop('annotations')
        annotations_per_type=defaultdict(list)

        for ann in annotations:
            annotations_per_type[ann.type].append([ann.span_start,ann.span_end,ann.features])

        state['annotations']=dict(annotations_per_type)
        return state #json.dumps(state)

    @classmethod
    def __from_minimized__(cls,state):

        all_annotations=[]
        for ann_type,annotations in state['annotations'].items():
            for ann in annotations:
                all_annotations.append(Annotation(ann[0], ann[1], ann_type, ann[2]))
        return cls(state['name'],state['text'],all_annotations,state['features'])


    def get_annotations_with_text(self, selector):
        """
        :param selector: textual string in one of the following formats:
           a) annotation_name
           b) annotation_name/feature_name       so you can do for instance stopword tagging on lemmas
        :return: list of selected (annotation,text) tuples
        """
        annotations_with_text = []
        selector_split = selector.split("/")
        element_annotation = selector_split[0].strip()
        element_feature = False if len(selector_split) == 1 else selector_split[1].strip()

        for a in self.annotations:
            if a.type == element_annotation:
                try:
                    if element_feature:
                            text = a.features[element_feature]
                    else:
                        text = self.text[a.span_start:a.span_end+1]
                    annotations_with_text.append((a, text))
                except KeyError:
                     #raise KeyError("The Annotation (%s) does not have feature named '%s'!" % (a.__str__(), element_feature))
                     pass
        return annotations_with_text
    def get_annotations(self, selector):
        return [a[0] for a in self.get_annotations_with_text(selector)]

    def get_annotation_texts(self,selector,stop_word_feature_name="StopWord"):
        return [text for (ann,text) in self.get_annotations_with_text(selector)
                               if not ann.features.has_key(stop_word_feature_name)]

    def raw_text(self,selector=None,stop_word_feature_name="StopWord",join_annotations_with=" "):
        if not selector:
            return self.text
        else:
            selected_subtexts=self.get_annotation_texts(selector,stop_word_feature_name)
            return join_annotations_with.join(selected_subtexts)

    def get_first_label(self,label_feature_name="Labels"):
        label_value=self.features.get(label_feature_name,None)
        #for klass in classes:
        #    if klass in self.features:
        #        return klass
        #d=444
        if label_value==None:
            return ''
        try:
            #print label_value
            true_value=json.loads(label_value)
            return true_value[0] if type(true_value)==list else label_value #if not a json list
        except ValueError, e:
            return label_value


import json
import urllib
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
import os.path
import time
import re
import zipfile
import document_converters
from tf_core.annotation import Annotation
from tf_core.document import Document
from tf_core.document_corpus import DocumentCorpus
from part_of_speech_tagging import corpus_reader


def load_adc(input_dict):
    """
    This widget processes text and loads it into ADC (Annotated Document Corpus) structure. The input file contains one
    document per line - the whole line represents text from the body of a document. In case lines contain more document
    properties (i.e.: ids, titles, labels,...) than other widgets should be used to load ADC structure.

    :param input: file path, string or list
    :param tab_separated_title: document title is separated from the text with \t. Example: title \t start of text
    :param leading_labels: documents has labels infront of the text. Example: !LB1 !Lb2 !LBL \t start of text
    :return adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    """
    input_text = input_dict[u"input"]
    tab_separated_title = input_dict['tab_separated_title'] == "true"
    leading_labels = input_dict['leading_labels'] == "true"

    docs, source, source_date,titles=_process_input(input_text,leading_labels)

    corpus_date = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))
    documents, labels = _process_adc(docs, tab_separated_title, leading_labels, titles)
    features = {u"Source": source, u"SourceDate": source_date, u"CorpusCreateDate": corpus_date, "Labels": json.dumps(labels)}

    return {"adc": DocumentCorpus(documents=documents, features=features)}


def load_ptb_corpus(input_dict):
    input_text = input_dict[u"input"]
    leading_labels = False
    docs, source, source_date,titles=_process_input(input_text,leading_labels)
    tagged_sents = []
    if "." in source:
        name = source.split('.')[0]
    else:
        name = source
    for doc in docs:
        l = []
        doc = doc.replace('\n', '')
        sent = ""
        depth = 0
        for character in doc:
            if character == '(':
                depth = depth + 1
            elif character == ')':
                depth = depth - 1
            sent += character
            if depth == 0 and len(sent) > 0 and not sent.isspace():
                l.append(sent)
                sent = ""
        for sent in l:
            tagged_sent = []
            match = re.findall(r'\(([^\(\)]*)\)',sent)
            for m in match:
                tagged_word = m.strip().split(" ")
                if len(tagged_word) == 2:
                    tagged_word = [tagged_word[1], tagged_word[0]]
                    tagged_sent.append(tagged_word)
            tagged_sents.append(tagged_sent)

    return {"ptb_corpus": [name, tagged_sents]}


def ptb_to_adc_converter(input_dict):
    corpus = input_dict['ptb_corpus'][1]
    annotation_feature = input_dict['annotation_name']
    annotations = []
    string_list = []
    docs = []
    title = u"Document1"
    features = {u"contentType": u"Text", u"sourceFileLine": '1'}
    position = 0
    for i, sentence in enumerate(corpus):
        if i%100 == 0 and i != 0 and i != (len(corpus) - 1):
            annotation = Annotation(0, position , "TextBlock")
            annotations.append(annotation)
            rawtext = " ".join(string_list)
            document = Document(title, rawtext, annotations, features)
            docs.append(document)
            annotations = []
            string_list = []
            title = u"Document" + str(i)
            features = {u"contentType": u"Text", u"sourceFileLine": '1'}
            position = 0

        sentence_start = position
        for word, tag in sentence:
            if len(word) > 0:
                string_list.append(word)
                word_length = len(word)
                annotation_features = {annotation_feature: tag}
                annotation = Annotation(position, position + word_length - 1, "Token", annotation_features)
                annotations.append(annotation)
                position = position + word_length + 1
        if position > sentence_start:
            annotation = Annotation(sentence_start, position - 1, "Sentence")
            annotations.append(annotation)
    annotation = Annotation(0, position , "TextBlock")
    annotations.append(annotation)
    rawtext = " ".join(string_list)
    document = Document(title, rawtext, annotations, features)
    docs.append(document)
    source = input_dict['ptb_corpus'][0]
    source_date = "unknown"
    corpus_date = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))
    features = {u"Source": source, u"SourceDate": source_date, u"CorpusCreateDate": corpus_date, "Labels": []}
    adc = DocumentCorpus(documents=docs, features=features)

    return {"adc": adc}


def nltk_corpus_to_adc(input_dict):
    chunk = input_dict['training_corpus']['chunk']
    corpus = input_dict['training_corpus']['corpus']
    annotation_feature = input_dict['annotation_name']
    name = "NLTK corpus"
    match = re.search(r"(\\\\|/)(\w+)'", str(corpus))
    if match:
        name = match.group(2) + " corpus"
    corpus_dict = {"ptb_corpus": [name, list(corpus_reader(corpus, chunk))], 'annotation_name': annotation_feature}
    return ptb_to_adc_converter(corpus_dict)


def crawl_url_links(input_dict):
    """
    This widget takes either a list of url links or a string, where each url is in a separate line. For every inputted url,
    the system crawls the page and extracts content using the readability library.

    :param urls: file path, string or list
    :param label: label which is set to all documents
    :return adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    """

    extractor_name=input_dict.get('extractor','DefaultExtractor')
    import requests
    label=input_dict['label']
    urls,_, source, source_date=_process_input(input_dict['input'],False)


    docs=[]
    titles=[]
    for url in urls:
        print url
        try:
            r = requests.get(url)
        except ConnectionError:
            continue
        if r.status_code==200:
            html=r.text
            from boilerpipe.extract import Extractor
            extractor = Extractor(extractor=extractor_name, html=html)

            titles.append(url)
            text=''
            if label:
                text+='!'+label+'\t'
            text+=extractor.getText()
            docs.append(text)


    corpus_date = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))
    documents, labels = _process_adc(docs, False, label, titles)
    features = {u"Source": source, u"SourceDate": source_date, u"CorpusCreateDate": corpus_date,
                "Labels": json.dumps([label]) if label else '[]'}

    return {"adc": DocumentCorpus(documents=documents, features=features)}


def search_with_faroo(input_dict):
    faroo_search="http://www.faroo.com/api?q={query}&start=1&length={length}&l=en&src=web&f=json&key={key}"
    file_name="workflows/nltoolkit/package_data/faroo_api_key.json"

    limit=int(input_dict.get('limit','50'))
    query=input_dict['query']

    if not query:
        raise StandardError("Please specify some search keywords.")

    urls=[]
    if os.path.isfile(file_name):
        with open(file_name) as data_file:
            api=json.load(data_file)
            query = '%27' + urllib.quote(query) + '%27'

            response=requests.get(faroo_search.format(key=api, query=query, length=limit))

            if response.status_code==200:
                results=json.loads(response.text)['results']
                for result in results:
                    urls.append(result['url'])
            else:
                raise StandardError(response.content())

    else:
        raise StandardError("Please create specify your Bing Api key in workflows/nltoolkit/package_data/faroo_api_key.json")

    return {'urls': urls}

def search_with_bing(input_dict):
    file_name="workflows/nltoolkit/package_data/bing_api_key.json"

    limit=int(input_dict.get('limit','50'))
    query=input_dict['query']

    if not query:
        raise StandardError("Please specify some search keywords.")

    urls=[]
    if os.path.isfile(file_name):
        with open(file_name) as data_file:
            api=json.load(data_file)
            query = '%27' + urllib.quote(query) + '%27'

            base_url = 'https://api.datamarket.azure.com/Bing/Search/Web'
            url = base_url + '?Query=' + query + '&$top=' + str(limit) + '&$format=json'

            # create credential for authentication
            user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36"

            auth = HTTPBasicAuth(api, api)
            headers = {'User-Agent': user_agent}

            response_data = requests.get(url, headers=headers, auth = auth)
            result_list = response_data.json()['d']['results']

            for result in result_list:
                #{u'Description': u'Reading the newest QubeGB reviews is a thing that could possibly be amazed for. This UK based communication firm gives a great choice of services to their customers.', u'Title': u'Kemi Anita Dasilva Ibru', u'Url': u'http://sfasf.com/', u'__metadata': {u'type': u'WebResult', u'uri': u"https://api.datamarket.azure.com/Data.ashx/Bing/Search/Web?Query='sfasf'&$skip=0&$top=1"}, u'DisplayUrl': u'sfasf.com', u'ID': u'b5811fb0-d21d-4868-a4ee-be32e3eae759'}
                urls.append(result['Url'])
    else:
        raise StandardError("Please create specify your Bing Api key in workflows/nltoolkit/package_data/bing_api_key.json")


    return {'urls': urls}


def load_mysql_document_corpus(input_dict):
    user = str(input_dict['user'])
    password = str(input_dict['password'])
    host = str(input_dict['host'])
    database_name = str(input_dict['database'])

    text_column_name = str(input_dict['text_column_name'])
    title_column_name = str(input_dict['title_column_name'])
    label_column_name = str(input_dict['label_column_name'])
    table_name = str(input_dict['table_name'])


    import MySQLdb
    db=MySQLdb.connect(passwd=password,user=user,db=database_name,host=host)
    c=db.cursor()
    c.execute("""SELECT %s, %s, %s FROM %s""" % (title_column_name,text_column_name,label_column_name,table_name))

    documents=[]
    labels=set()
    for (title, text, label) in c:
        doc_labels=[t.rstrip() for t in label.split(",")]
        labels.update(doc_labels)
        features={"Labels": json.dumps(doc_labels) }
        documents.append(Document(name=title,
                              features=features,
                              text=unicode(text),
                              annotations=[Annotation(span_start=0,span_end=max(0, len(unicode(text)) - 1),
                                                      type=u"TextBlock", features={})]))
    c.close()
    features = {u"Source": 'MySQL DB: '+database_name, u"CorpusCreateDate": unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime())),
                "Labels": json.dumps(list(labels)) }
    return {'adc': DocumentCorpus(documents=documents, features=features)}



def _process_input(input_text,tab_separated_title):
    source = "list"
    source_date = "unknown"
    titles = []


    #check if input is a file
    if type(input_text) != list and os.path.exists(input_text):
        source = os.path.basename(input_text)
        seconds = os.path.getctime(input_text)
        source_date = unicode(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime(seconds)))
        ext = os.path.splitext(input_text)[1]

        if ext == ".zip":
            f = zipfile.ZipFile(input_text, "r")
            dir = os.path.dirname(input_text)
            f.extractall(dir)
            input_text = []

            if len(f.namelist()) == 0:
                raise Exception("Zip file is empty!")

            for name in f.namelist():
                #OSX can have a journaling file in it or the file is a directory
                if name not in ["._.DS_Store", ".DS_Store", "__MACOSX"] and name[-1] != "/":
                    filename, text = document_converters.document_to_text(dir + os.sep + name)
                    input_text.append(text)
                    if not tab_separated_title:
                        titles.append(filename)
        else:
            filename, input_text = document_converters.document_to_text(input_text)

    #check if input is a string
    if type(input_text) == unicode or type(input_text) == str:
        source = "string" if source == "list" else source
        input_text = re.split("[\r\n]", input_text)

    return input_text, source, source_date,titles

def _process_adc(texts,  tab_separated_title, leading_labels, titles=[]):
    """
    function parses title and labels for documents from text and it sets annnotation for each document.

    :param texts: list of documents
    :param tab_separated_title: document title is separated from the text with \t. Example: title \t start of text
    :param leading_labels: documents has labels infront of the text. Example: !LB1 !Lb2 !LBL \t start of text
    :return: list of documents, uniq list of labels
    """
    documents = []
    corpus_labels = set()
    for i, text in enumerate(texts):
        if text:
            title = u"Document" + unicode(i + 1) if titles == [] else titles[i]
            features = {u"contentType": u"Text", u"sourceFileLine": unicode(i)}
    
            if tab_separated_title:
                #example: title \t start of text
                text = text.split("\t")
                title = unicode(text[0])
                text = "\t".join(text[1:])
    
            if leading_labels:
                #example: !LB1 !Lb2 !LBL \t start of text
                text = text.split("\t")
                doc_labels=[]
                for label in [f.strip() for f in text[0].split("!") if f != u""]:
                    features[label] = "true"
                    corpus_labels.add(label)
                    doc_labels.append(label)
                text = "".join(text[1:])
                features["Labels"]=json.dumps(doc_labels)
    
            documents.append(Document(name=title,
                                      features=features,
                                      text=unicode(text),
                                      annotations=[Annotation(span_start=0,
                                                              span_end=max(0, len(unicode(text)) - 1),
                                                              type=u"TextBlock",
                                                              features={})]))
    return documents, list(corpus_labels)


def get_plain_texts(input_dict):
    """
    Widget transforms Annotated Document Corpus to string.

    :param adc: Annotated Document Corpus.
    :param feature_annotation: Select a feature annotation.
    :param delimiter: Delimiter for token concatenation.
    :param include_doc_id: Include Document Identifier.
    :return: String with all documents in Annotated Document Corpus.

    """

    adc = input_dict['adc']
    annotation = input_dict['token_annotation']
    feature_annotation = input_dict['feature_annotation']
    if feature_annotation:
        annotation+="/"+feature_annotation
    delimiter = input_dict['delimiter']
    includeDocId = input_dict['include_doc_id'] == "true"

    output_dict = {"strings": []}
    for document in adc.documents:
        output_dict["strings"].append((document.name+": " if includeDocId else "") + delimiter.join([t[1] for t in document.get_annotations_with_text(annotation)]))
    return output_dict


def display_document_corpus(input_dict):
    #implemented in visualization_views.py
    return {}

def display_annotation_statistic(input_dict):
    #implemented in visualization_views.py
    return {}


import copy
from itertools import izip
import json
import re
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer, TfidfTransformer
import nltk
from collections import defaultdict
from sklearn.tree import DecisionTreeClassifier
from sklearn.naive_bayes import GaussianNB, MultinomialNB





def simulate_cf_pickling(obj_to_pickle,compress_object=False):
    from base64 import b64encode, b64decode
    from zlib import compress, decompress
    from cPickle import dumps,loads

    if not compress_object:
        return loads(b64decode(b64encode(dumps(obj_to_pickle))))
    else:
        return loads(decompress(b64decode(b64encode(compress(dumps(obj_to_pickle))))))

#python manage.py export_package workflows/nltoolkit/db/package_data.json nltoolkit
#python manage.py export_package workflows/literature_based_discovery/db/package_data.json literature_based_discovery
#python manage.py celery worker -l info



if __name__=="__main__":
    """quick test, pickling and depickling"""
    from cPickle import dumps,loads
    from base64 import b64encode, b64decode

    dc=DocumentCorpus([Document("name","text",[Annotation(1,2,'token',{'stopword': True})],{})],{'created_at':'now'})
    dc2=loads(b64decode(b64encode(dumps(dc))))



from tf_core.bow_dataset import BowDataset
from tf_core.bow_model_contructor import BowModelConstructor


def construct_bow_model_constructor(input_dict):
    adc=input_dict['adc']
    token_annotation = input_dict['token_annotation']
    stem_feature_name = input_dict['stem_feature_name']

    if stem_feature_name:
        token_annotation+='/'+stem_feature_name

    args={  'adc': adc,
            'token_annotation': token_annotation,
            'stem_feature_name':stem_feature_name,
            'stop_word_feature_name': input_dict['stop_word_feature_name'],
            'label_doc_feature_name': input_dict['label_doc_feature_name'],
            'weighting_type':input_dict['word_weight_type'], #tf_idf
            'normalize_vectors': input_dict['normalize_vectors']=='true', #true
            'max_ngram': int(input_dict['max_ngram_len']), #2
            'min_tf': int(input_dict['min_word_freq']),#5
            'predefined_vocabulary': input_dict.get('vocabulary',None)}

    bow_model=BowModelConstructor(**args)

    return {'bow_model_constructor': bow_model}

def construct_dataset_and_bow_model_constructor(input_dict):
    """
    In TextFlows the Construct BoW Dataset and BoW Model Constructor widget takes as an input an ADC data object
    and generates a sparse BoW model dataset (which can be then handed to i.e. a classifier). The widget takes as an
    input also several user defined parameters, which are taken into account when building the feature dataset:

    :param Token Annotation: This is the type of Annotation instances, which mark parts of the document (e.g., words, sentences or a terms), which will be used for generating the vocabulary and the dataset.
    :param Feature Name: If present, the model will be constructed out of annotations' feature values instead of document text. For example, this is useful when we wish build the BoW model using stems instead of the original word forms.
    :param Stop Word Feature Name: This is an annotation feature name which was used to tag tokens as stop words. These tokens will be excluded from the BoW representational model. If blank, no stop words will be used.
    :param Label Document Feature Name: This is the name of the document feature which will be used for class labeling examples in the dataset. If blank, the generated sparse dataset will be unlabeled.
    :param Maximum n-gram Length: The upper boundary of the range of n-values for different n-grams to be extracted. All values of $n$ such that $1 \leq n \leq \mbox{max\_ngram$ will be used.
    :param Minimum Word Frequency: When building the vocabulary ignore terms that have a term frequency strictly lower than the given threshold. This value is also called cut-off in the literature.
    :param Word Weighting Type: The user can select among various weighting models for assigning weights to features:

        Binary. A feature weight is 1 if the corresponding term is present in the document, or zero otherwise.
        Term occurrence. A feature weight is equal to the number of occurrences of the corresponding term. This weight might be sometimes better than a simple binary value since frequently occurring term are likely to be more relevant to the given text.
        Term frequency. A weight is derived from the term occurrence by dividing the vector by the sum of all vector's weights. The reasoning is similar to term occurrence, with normalization with respect to document size.
        TF-IDF. Term Frequency-Inverse Document Frequency~\citep{salton1988term is the most common scheme for weighting features. For a given term $w$ in document $d$ from corpus $D$, the TF-IDF measure is defined as follows:
        Safe TF-IDF. For a given term $w$ in document $d$ from corpus $D$, the Safe TF-IDF measure is defined as follows:
        TF-IDF with sublinear TF scaling. It often seems unlikely that twenty occurrences of a term in a document truly carry twenty times the significance of a single occurrence. Accordingly, there has been considerable research into variants of term frequency that go beyond counting the number of occurrences of a term~\citep{paltoglou2010study. A common modification is to use the logarithm of the term frequency instead of tf, defined as:

    :param Normalize Vectors: The weighting methods can be further modified by vector normalization. If the user opts to use it in TextFlows the $L2$ regularization~\citep{ng2004feature is performed.

    :return:  Besides the sparse BoW model dataset the Construct BoW Dataset and BoW Model Constructor also outputs
     a BowModelConstructor instance. This additional object contains settings which allow repetition of the feature
     construction steps on another document corpus. These settings include the inputted parameters, as well as the
     learned term weights and vocabulary.

    """

    adc=input_dict['adc']

    bow_model=construct_bow_model_constructor(input_dict)['bow_model_constructor']
    bow_dataset=BowDataset.from_adc(adc,bow_model)

    return {'bow_model_constructor': bow_model,'bow_dataset': bow_dataset}

def create_dataset_using_bow_model_constructor(input_dict):
    bow_model = input_dict['bow_model_constructor']
    adc = input_dict['adc']

    return {'bow_dataset': BowDataset.from_adc(adc,bow_model)}


from itertools import izip
import inspect

def universal_word_tagger_hub(adc,tagger_dict,input_annotation,output_annotation, pos_annotation=None):
    tagger=tagger_dict['object']
    tagger_function=tagger_dict['function']
    args=tagger_dict.get('args',[])
    kwargs=tagger_dict.get('kargs',{})

    for document in adc.documents:
        if document.features['contentType'] == "Text":
            if not document.text:
                pass
            for annotation,subtext in document.get_annotations_with_text(input_annotation): #all annotations of this type
                if subtext:
                    if pos_annotation:
                        if pos_annotation in annotation.features:
                            method = getattr(tagger,tagger_function)
                            if inspect.getargspec(method)[2] == 'kwargs':
                                kwargs[pos_annotation] = annotation.features[pos_annotation]
                    new_feature=getattr(tagger,tagger_function)(subtext,*args,**kwargs)
                    if new_feature!=None:
                        annotation.features[output_annotation]=new_feature
    return {'adc': adc }


def universal_sentence_tagger_hub(input_dict):
    tagger_dict = input_dict['pos_tagger']
    tagger=tagger_dict['object']
    tagger_function=tagger_dict['function']
    args=tagger_dict.get('args',[])
    kwargs=tagger_dict.get('kargs',{})

    group_annotation_name = input_dict['group_annotation']
    element_annotation_name = input_dict['element_annotation']
    output_annotation_name = input_dict['output_feature']
    adc = input_dict['adc']


    for doc in adc.documents:
        if doc.features['contentType'] == "Text":
            if not doc.text:
                pass
            group_annotations=sorted(doc.get_annotations_with_text(group_annotation_name),key=lambda x: x[0].span_start)
            element_annotations=sorted(doc.get_annotations_with_text(element_annotation_name),key=lambda x: x[0].span_start)

            text_grouped=[] #text_groups= [['First','sentence',['Second','sentance']]
            annotations_grouped=[] #annotations_grouped= [[<Annotation span_start:0 span_ned:4>, <Annotation span_start:6 span_ned:11>],[...

            i=0
            for group_annotation,_ in group_annotations:
                elements=[]
                sentence_annotations=[]
                #find elementary annotations 'contained' in the group_annotation
                while i<len(element_annotations) and element_annotations[i][0].span_end<=group_annotation.span_end:
                    annotation=element_annotations[i][0]
                    text_block=element_annotations[i][1]
                    elements.append(text_block)
                    sentence_annotations.append(annotation)
                    i+=1
                text_grouped.append(elements)
                annotations_grouped.append(sentence_annotations)

            
            new_features=getattr(tagger,tagger_function)(text_grouped,*args,**kwargs)

            for sentence_features, sentence_annotations in izip(new_features,annotations_grouped):
                for feature,annotation in izip(sentence_features,sentence_annotations):
                    annotation.features[output_annotation_name]=feature[1] #[0:number_of_letters]

    return {'adc': adc }


from itertools import izip

import nltk
from tagging_common import universal_word_tagger_hub
#from tagging_common_parallel import universal_word_tagger_hub


def chunking_hub(input_dict):
    adc = input_dict['adc']
    chunker_dict = input_dict['chunker']
    chunker = input_dict['chunker']['object']
    chunker_function = input_dict['chunker']['function']
    output_annotation = input_dict['output_feature']
    input_annotation = input_dict['input_feature']
    args = chunker_dict.get('args',[])
    kwargs = chunker_dict.get('kargs',{})

    for doc in adc.documents:
        if doc.features['contentType'] == "Text":
            if not doc.text:
                pass
            group_annotations=sorted(doc.get_annotations_with_text("Sentence"),key=lambda x: x[0][0])
            element_annotations=sorted(doc.get_annotations_with_text('Token'),key=lambda x: x[0][0])

            text_grouped=[] #text_groups= [['First','sentence',['Second','sentance']]
            annotations_grouped=[] #annotations_grouped= [[<Annotation span_start:0 span_ned:4>, <Annotation span_start:6 span_ned:11>],[...

            i=0
            for group_annotation,_ in group_annotations:
                sentence_annotations=[]
                #find elementary annotations 'contained' in the group_annotation
                while i<len(element_annotations) and element_annotations[i][0][1]<=group_annotation[1]:
                    annotation_list=element_annotations[i][0][3]
                    annotation = ""
                    for name, value in annotation_list:
                        if name == input_annotation:
                            annotation = value
                            break
                    text_block=element_annotations[i][1]
                    sentence_annotations.append((text_block, annotation))
                    i+=1
                annotations_grouped.append(sentence_annotations)
                #print list(izip(annotations_grouped, text_grouped))
            new_features=[getattr(chunker,chunker_function)(tagged_sentence,*args,**kwargs) for tagged_sentence in annotations_grouped]
            for annotation, chunk in izip(group_annotations, new_features):
                annotation[0][3].append((output_annotation, chunk)) #[0:number_of_letters]

    return {'adc': adc }

# Chunkers
def regex_parser(input_dict):
    grammar = input_dict['grammar']
    """
    regex parser, requires grammar as input
    """
    return {'chunker':
                {'object': nltk.RegexpParser(grammar),
                 'function': 'parse', }
    }

def tags_since_dt(sentence, i):
    tags = set()
    for word, pos in sentence[:i]:
        if pos == 'DT':
            tags = set()
        else:
            tags.add(pos)
    return '+'.join(sorted(tags))


def npchunk_features(sentence, i, history):
    word, pos = sentence[i]
    if i == 0:
        prevword, prevpos = "<START>", "<START>"
    else:
        prevword, prevpos = sentence[i-1]
    if i == len(sentence)-1:
        nextword, nextpos = "<END>", "<END>"
    else:
        nextword, nextpos = sentence[i+1]
    return {"pos": pos,
            "word": word,
            "prevpos": prevpos,
            "nextpos": nextpos,
            "prevpos+pos": "%s+%s" % (prevpos, pos),
            "pos+nextpos": "%s+%s" % (pos, nextpos),
            "tags-since-dt": tags_since_dt(sentence, i)}


class ConsecutiveNPChunkTagger(nltk.TaggerI):

    def __init__(self, train_sents):
        train_set = []
        for tagged_sent in train_sents:
            untagged_sent = nltk.tag.untag(tagged_sent)
            history = []
            for i, (word, tag) in enumerate(tagged_sent):
                featureset = npchunk_features(untagged_sent, i, history)
                train_set.append( (featureset, tag))
                history.append(tag)
        self.classifier = nltk.NaiveBayesClassifier.train(train_set)

    def tag(self, sentence):
        history = []
        for i, word in enumerate(sentence):
            featureset = npchunk_features(sentence, i, history)
            tag = self.classifier.classify(featureset)
            history.append(tag)
        return zip(sentence, history)


class ClassifierBasedChunkParser(nltk.ChunkParserI):
    def __init__(self, train_sents):
        tagged_sents = [[((w,t),c) for (w,t,c) in nltk.chunk.tree2conlltags(sent)] for sent in train_sents]
        self.tagger = ConsecutiveNPChunkTagger(tagged_sents)

    def parse(self, sentence):
        tagged_sents = self.tagger.tag(sentence)
        conlltags = [(w,t,c) for ((w,t),c) in tagged_sents]
        return nltk.chunk.conlltags2tree(conlltags)


from nltk.corpus import conll2000

def classifier_based_parser(input_dict):
    train_sents = conll2000.chunked_sents('train.txt', chunk_types=['NP'])
    classifier_based_chunker = ClassifierBasedChunkParser(train_sents)
    return {'chunker':
                {'object': classifier_based_chunker,
                 'function': 'parse', }
    }


import multiprocessing
from functools import partial
from itertools import izip


def tag_document(document,tagger,tagger_function,args,kwargs,input_annotation,output_annotation):
    if document.features['contentType'] == "Text":
        if not document.text:
            pass
        for annotation,subtext in document.get_annotations_with_text(input_annotation): #all annotations of this type
            if subtext:
                new_feature=getattr(tagger,tagger_function)(subtext,*args,**kwargs)
                if new_feature!=None:
                    annotation.features[output_annotation]=new_feature
    return document

def universal_word_tagger_hub(adc,tagger_dict,input_annotation,output_annotation):
    tagger=tagger_dict['object']
    tagger_function=tagger_dict['function']
    args=tagger_dict.get('args',[])
    kwargs=tagger_dict.get('kargs',{})

    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())

    print "evo nas!!!"
    #parallel for document in adc.documents:
    new_documents=pool.map(
        partial(tag_document,
                tagger=tagger,
                tagger_function=tagger_function,
                args=args,
                kwargs=kwargs,
                input_annotation=input_annotation,
                output_annotation=output_annotation),
        adc.documents,
        100 #chunksize, constructs list of this size which are passed to pool workers
    )
    pool.close()
    pool.join()
    adc.documents=new_documents #list(new_documents)

    print "dijo!2!!"

    return {'adc': adc }

def sentance_tag_a_document(doc,tagger,tagger_function,args,kwargs,
                            element_annotation_name,group_annotation_name,output_annotation_name):
    if doc.features['contentType'] == "Text":
        if not doc.text:
            pass
        group_annotations=sorted(doc.get_annotations_with_text(group_annotation_name),key=lambda x: x[0].span_start)
        element_annotations=sorted(doc.get_annotations_with_text(element_annotation_name),key=lambda x: x[0].span_start)

        text_grouped=[] #text_groups= [['First','sentence',['Second','sentance']]
        annotations_grouped=[] #annotations_grouped= [[<Annotation span_start:0 span_ned:4>, <Annotation span_start:6 span_ned:11>],[...

        i=0
        for group_annotation,_ in group_annotations:
            elements=[]
            sentence_annotations=[]
            #find elementary annotations 'contained' in the group_annotation
            while i<len(element_annotations) and element_annotations[i][0].span_end<=group_annotation.span_end:
                annotation=element_annotations[i][0]
                text_block=element_annotations[i][1]
                elements.append(text_block)
                sentence_annotations.append(annotation)
                i+=1
            text_grouped.append(elements)
            annotations_grouped.append(sentence_annotations)

        new_features=getattr(tagger,tagger_function)(text_grouped,*args,**kwargs)
        for sentence_features, sentence_annotations in izip(new_features,annotations_grouped):
            for feature,annotation in izip(sentence_features,sentence_annotations):
                annotation.features[output_annotation_name]=feature[1]
    return doc



def universal_sentence_tagger_hub(input_dict):
    tagger_dict = input_dict['pos_tagger']
    tagger=tagger_dict['object']
    tagger_function=tagger_dict['function']
    args=tagger_dict.get('args',[])
    kwargs=tagger_dict.get('kargs',{})

    group_annotation_name = input_dict['group_annotation']
    element_annotation_name = input_dict['element_annotation']
    output_annotation_name = input_dict['output_feature']
    adc = input_dict['adc']


    pool = multiprocessing.Pool(processes=multiprocessing.cpu_count())

    print "evo nas!!!"
    #parallel for document in adc.documents:
    new_documents=pool.map(
        partial(sentance_tag_a_document,
                tagger=tagger,
                tagger_function=tagger_function,
                args=args,
                kwargs=kwargs,
                element_annotation_name=element_annotation_name,
                group_annotation_name=group_annotation_name,
                output_annotation_name=output_annotation_name),
        adc.documents,
        100 #chunksize, constructs list of this size which are passed to pool workers
    )
    pool.close()
    pool.join()
    adc.documents=new_documents
    print "dijo!!!"
    return {'adc': adc }




# def chunks(l, n):
#     c=[[] for _ in range(n)]
#     for i in range(l):
#         c[i%n].append(i)
#     return c
#
# print chunks(10,6)
#
# from multiprocessing import Process, Value, Array, Pool
#
# def f(a,indices):
#     for i in indices:
#         a[i] = -a[i]
#
# if __name__ == '__main__':
#     a=[[i] for i in range(100)]
#     arr = Array('i', a)
#
#     no_of_workers=6
#     workers=[Process(target=f, args=(arr, indices)) for indices in chunks(len(arr),no_of_workers)]
#
#
#     for p in workers:
#         p.start()
#     for p in workers:
#         p.join()
#
#     print arr[:]
#     print a
#
#
#     #pool = multiprocessing.Pool(processes=6)
#     #case_data = RAW_DATASET
#     #pool.apply(f, args=(num, arr))
#     #pool.close()
#     #pool.join()



import re

import nltk


class TokenSplitter(object):
    '''Used for splitting tokens in BoW construction.'''
    regex_parser=None

    def __init__(self):
        self.regex_parser=re.compile("^[_A-Za-z0-9']+$")
    def __call__(self, doc):
        if doc.startswith("Substance P"):
            sdfasdfa=333
        return [d for d in doc.split('|##|') if self.regex_parser.match(d)]

class NltkCorpus():
    """ Wrapper for Nltk corpora. In Nltk 3.x Nltk corpora is not picklable.
    """
    corpus_name=""
    corpus=None
    _corpus_methods=[]
    def __init__(self,name):
        self.corpus_name=name
        self._corpus_methods=dir(getattr(nltk.corpus,self.corpus_name))

    def _corpus(self):
        self.corpus = self.corpus or getattr(nltk.corpus,self.corpus_name)
        return self.corpus

    def __getattr__(self, name):
        if not name in self._corpus_methods:
            raise AttributeError
        else:
            def method():
                return getattr(self._corpus(),name)()
            return method

    def __repr__(self):
        return "NltkCorpus wrapper for "+self.corpus_name+" dataset"

    def __getstate__(self):
        return {'corpus_name': self.corpus_name,'_corpus_methods': self._corpus_methods}


class NltkRegexpTokenizer():
    """ Wrapper for Nltk RegexTokenizer. Python's regular expressions are not picklable.
    """
    _pattern=''
    _kargs={}

    def __init__(self,pattern,**kargs):
        self._pattern=pattern
        self._kargs=kargs

    def span_tokenize(self,text):
        return nltk.RegexpTokenizer(self._pattern,**self._kargs).span_tokenize(text)

class NltkRegexpStemmer():
    """ Wrapper for Nltk RegexStemmer. Python's regular expressions are not picklable.
    """
    _pattern=''
    _kargs={}

    def __init__(self,pattern,**kargs):
        self._pattern=pattern
        self._kargs=kargs

    def stem(self,text):
        return nltk.stem.RegexpStemmer(self._pattern,**self._kargs).stem(text)


class NltkClassifier():
    """ This is a wrapper for Nltk classifiers. Nltk classifiers do not have an appropriate __init__
        method which could save kargs and use them latter when .train(train_data) is called.
    """
    _classifier=None
    _kargs={}

    def __init__(self,csf,**kargs):
        self._classifier=csf
        self._kargs=kargs

    def train(self,training_data):
        self._classifier=self._classifier.train(training_data,**self._kargs)
        return self

    def prob_classify_many(self,testing_dataset):
        return self._classifier.prob_classify_many(testing_dataset)

    def __repr__(self):
        return '<NltkClassifier object for %s>' % (self._classifier)

from nltk.probability import DictionaryProbDist

class DictionaryProbDist(DictionaryProbDist):
    def __repr__(self):
        return '<ProbDist: %s>' % str(self._prob_dict)

    @classmethod
    def from_prediction_and_classes(cls,prediction,classes):
        prob_dict={}
        for i,klass in enumerate(classes):
            prob_dict[unicode(klass)]=1. if klass==prediction else 0.
        return cls(prob_dict=prob_dict)

    @classmethod
    def from_probabilities_and_classes(cls,predictions,classes):
        prob_dict={}
        for i,klass in enumerate(classes):
            prob_dict[unicode(klass)]=predictions[i]
        return cls(prob_dict=prob_dict)




from nltk.tokenize.stanford import StanfordTokenizer
from nltk.tokenize import TextTilingTokenizer
from nltk.tokenize import TreebankWordTokenizer


class SpanTokenizeMixin():

    def span_tokenize(self, s):
        r"""
        Return the offsets of the tokens in *s*, as a sequence of ``(start, end)``
        tuples, by splitting the string at each successive match of *regexp*.

            #>>> from workflows.textflows import StanfordTokenizer
            #>>> s = '''Good muffins cost $3.88\nin New York.  Please buy me
            #... two of them.\n\nThanks.'''
            #>>> list(StanfordTokenizer().span_tokenize(s))
            #[(0, 4), (5, 12), (13, 17), (18, 23), (24, 26), (27, 30), (31, 36),
            #(38, 44), (45, 48), (49, 51), (52, 55), (56, 58), (59, 64), (66, 73)]

        :param s: the string to be tokenized
        :type s: str
        :rtype: iter(tuple(int, int))
        """
        tokens = self.tokenize(s)
        right = 0

        for token in tokens:
            offset=s[right:].find(token)
            if offset<0:
                sadfa=45
            else:
                #raise Exception('Negative offset!')
                left = right + offset #s[right:].find(token)
                right = left + len(token)
                yield left, right


class StanfordTokenizer(SpanTokenizeMixin, StanfordTokenizer):
    pass


class TextTilingTokenizer(SpanTokenizeMixin, TextTilingTokenizer):
    pass


class TreebankWordTokenizer(SpanTokenizeMixin, TreebankWordTokenizer):
    pass







class Annotation:
    def __init__(self, span_start, span_end, type, features=None):
        self.features=features or {}
        self.span_start=span_start
        self.span_end=span_end
        self.type=type

        if span_start<0 or span_end<0 or span_start>span_end:
            raise Exception('Invalid span_start or span_end in %s' % str(self))
        
    def __repr__(self):
        return '<Annotation span_start:%d span_ned:%d>' % (self.span_start, self.span_end)

    def __unicode__(self):
        return 'span_start: %d, span_ned: %d' % (self.span_start, self.span_end)
    
    def __str__(self):
        return unicode(self).encode('utf-8')


import nltk
from django.conf import settings

from tf_core.annotation import Annotation
from tf_core.nltoolkit.helpers import TreebankWordTokenizer, NltkRegexpTokenizer, TextTilingTokenizer, \
    StanfordTokenizer


def tokenizer_hub(input_dict):
    """
    Apply the *tokenizer* object on the Annotated Document Corpus (*adc*):

    1. first select only annotations of type *input_annotation*,
    2. apply the tokenizer
    3. create new annotations *output_annotation* with the outputs of the tokenizer.

    :param adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    :param tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    :param input_annotation: Which annotated part of document to be splitted.
    :param output_annotation: How to annotate the newly discovered tokens.

    :returns adc: Annotated Document Corpus (workflows.textflows.DocumentCorpus)
    """
    tokenizer_dict = input_dict['tokenizer']

    if type(tokenizer_dict)!=dict:
        from workflows.tasks import executeFunction
        from tf_latino.latino.library_gen import latino_tokenize_words

        return latino_tokenize_words(input_dict) if not settings.USE_WINDOWS_QUEUE \
            else executeFunction.apply_async([latino_tokenize_words,input_dict],queue="windows").wait()
    else:
        tokenizer=tokenizer_dict['object']
        args=tokenizer_dict.get('args',[])
        kwargs=tokenizer_dict.get('kargs',{})
        input_annotation = input_dict['input_annotation']
        output_annotation = input_dict['output_annotation']
        adc = input_dict['adc']
        docs_count=len(adc.documents)
        for i,document in enumerate(adc.documents):
            if document.features['contentType'] == "Text":
                if not document.text:
                    pass
                for annotation,subtext in document.get_annotations_with_text(input_annotation): #all annotations of this type
                    new_token_spans=tokenizer.span_tokenize(subtext,*args,**kwargs)
                    for starts_at,ends_at in new_token_spans:
                        document.annotations.append(Annotation(annotation.span_start+starts_at,annotation.span_start+ends_at-1,output_annotation))
            if i%100==0:
                print int((i+1)*1.0/docs_count*100)
            #widget.progress = int((i+1)*1.0/*100)
            #widget.save()
        return {'adc': adc}


def nltk_treebank_word_tokenizer(input_dict):
    """
    The Treebank tokenizer uses regular expressions to tokenize text as in Penn Treebank.
    This is the method that is invoked by ``word_tokenize()``.  It assumes that the
    text has already been segmented into sentences, e.g. using ``sent_tokenize()``.

    This tokenizer performs the following steps:
    - split standard contractions, e.g. ``don't`` -> ``do n't`` and ``they'll`` -> ``they 'll``
    - treat most punctuation characters as separate tokens
    - split off commas and single quotes, when followed by whitespace
    - separate periods that appear at the end of line

    :param input_dict (default): {}
    :returns tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """


    return {'tokenizer': {'object': TreebankWordTokenizer()}}


def nltk_punkt_sentence_tokenizer(input_dict):
    """
    A sentence tokenizer which uses an unsupervised algorithm to build
    a model for abbreviation words, collocations, and words that start
    sentences; and then uses that model to find sentence boundaries.
    This approach has been shown to work well for many European
    languages.

    :param input_dict (default): {}
    :returns tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    return {'tokenizer': {'object': nltk.PunktSentenceTokenizer()}}


def nltk_regex_tokenizer(input_dict):
    """
    The Regex Tokenizer splits a string into substrings using a regular expression.

    :param :param pattern: The pattern used to build this tokenizer.
        (This pattern may safely contain capturing parentheses.)
    :param gaps: True if this tokenizer's pattern should be used
        to find separators between tokens; False if this
        tokenizer's pattern should be used to find the tokens
        themselves.
    :param discard_empty: True if any empty tokens `''`
        generated by the tokenizer should be discarded.  Empty
        tokens can only be generated if `_gaps == True`.

    :param input_dict (default): {u'pattern': u'\\p{L}+(-\\p{L}+)*', u'gaps': u'', u'discard_empty': u''}
    :return: tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    pattern = input_dict[u'pattern']
    gaps = input_dict[u'gaps'] == "true"
    discard_empty = input_dict[u'discard_empty'] == "true"

    return {'tokenizer': {'object': NltkRegexpTokenizer(pattern, gaps=gaps, discard_empty=discard_empty)}}


def nltk_sexpression_tokenizer(input_dict):
    """
    A tokenizer that divides strings into s-expressions.
    An s-expresion can be either:
      - a parenthesized expression, including any nested parenthesized
        expressions, or
      - a sequence of non-whitespace non-parenthesis characters.

    :param input_dict (default): {u'parens': u'()', u'strict': u'true'}
    :param parens: A two-element sequence specifying the open and close parentheses
        that should be used to find sexprs.  This will typically be either a
        two-character string, or a list of two strings.
    :param strict: If true, then raise an exception when tokenizing an ill-formed sexpr.

    :return: tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    parens = input_dict[u'parens']
    strict = input_dict[u'strict'] == "true"

    return {'tokenizer': {'object': nltk.SExprTokenizer(parens=parens, strict=strict)}}


def nltk_simple_tokenizer(input_dict):
    """
    These tokenizers divide strings into substrings using the string
    ``split()`` method.
    When tokenizing using a particular delimiter string, use
    the string ``split()`` method directly, as this is more efficient.

    Space Tokenizer - Tokenize a string using the space character as a delimiter, which is the same as s.split(' ').
    Tab Tokenizer - Tokenize a string use the tab character as a delimiter, the same as s.split('\t').
    Char Tokenizer - Tokenize a string into individual characters.
    Whitespace Tokenizer - Tokenize a string on whitespace (space, tab, newline).
    Blankline Tokenizer - Tokenize a string, treating any sequence of blank lines as a delimiter. Blank lines are defined as lines containing no characters, except for space or tab characters.
    Word Punct Tokenizer - Tokenize a text into a sequence of alphabetic and non-alphabetic characters, using the regexp ``\w+|[^\w\s]+``.

    :param input_dict (default): {u'type': u'wordpunct_tokenizer'}
    :param (type unicode) u"char_tokenizer"
    :param (type unicode) u"space_tokenizer"
    :param (type unicode) u"tab_tokenizer"
    :param (type unicode) u"whitespace_tokenizer"
    :param (type unicode) u"blankline_tokenizer"
    :param (type unicode) u"wordpunct__tokenizer"

    :return: tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    if input_dict["type"] == u"char_tokenizer":
        tokenizer = nltk.tokenize.simple.CharTokenizer()
    elif input_dict["type"] == u"space_tokenizer":
        tokenizer = nltk.SpaceTokenizer()
    elif input_dict["type"] == u"tab_tokenizer":
        tokenizer = nltk.TabTokenizer()
    elif input_dict["type"] == u"whitespace_tokenizer":
        tokenizer = nltk.WhitespaceTokenizer()
    elif input_dict["type"] == u"blankline_tokenizer":
        tokenizer = nltk.BlanklineTokenizer()
    elif input_dict["type"] == u"wordpunct_tokenizer":
        tokenizer = nltk.WordPunctTokenizer()

    return {'tokenizer': {'object': tokenizer}}


def nltk_line_tokenizer(input_dict):
    """
    Tokenize a string into its lines, optionally discarding blank lines.
    This is similar to ``s.split('\n')``

    :param input_dict (default): {u'blanklines': u'discard'}
    :param blanklines: 'discard', 'keep', 'discard-eof'
    :return: tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    blanklines = input_dict[u"blanklines"]
    return {'tokenizer': {'object': nltk.LineTokenizer(blanklines=blanklines)}}


def nltk_stanford_tokenizer(input_dict):
    """
    Stanford Tokenizer

    :param input_dict (default): {}
    :return: tokenizer: A python dictionary containing the Tokenizer object and its arguments.
    """

    return {'tokenizer': {'object': StanfordTokenizer()}}


def nltk_text_tiling_tokenizer(input_dict):
    """
    Tokenize a document into topical sections using the TextTiling algorithm.
    This algorithm detects subtopic shifts based on the analysis of lexical
    co-occurrence patterns.


    :param input_dict (default): {u'smoothing_rounds': u'1', u'k': u'10', u'smoothing_width': u'2', u'stopwords': u'None', u'similarity_method': u'BLOCK_COMPARISON', u'w': u'20', u'cutoff_policy': u'HC'}
    :param w: Pseudosentence size
    :type w: int
    :param k: Size (in sentences) of the block used in the block comparison method
    :type k: int
    :param similarity_method: The method used for determining similarity scores:
       `BLOCK_COMPARISON` (default) or `VOCABULARY_INTRODUCTION`.
    :type similarity_method: constant
    :param stopwords: A list of stopwords that are filtered out (defaults to NLTK's stopwords corpus)
    :type stopwords: list(str)
    :param smoothing_width: The width of the window used by the smoothing method
    :type smoothing_width: int
    :param smoothing_rounds: The number of smoothing passes
    :type smoothing_rounds: int
    :param cutoff_policy: The policy used to determine the number of boundaries:
      `HC` (default) or `LC`
    :type cutoff_policy: constant
    """

    w = int(input_dict["w"])
    k = int(input_dict["k"])
    similarity_method = input_dict["similarity_method"]
    stopwords = None if input_dict["stopwords"] == "None" else [word.strip() for word in input_dict["stopwords"].split(",")]
    smoothing_width = int(input_dict["smoothing_width"])
    smoothing_rounds = int(input_dict["smoothing_rounds"])
    cutoff_policy = input_dict["cutoff_policy"]

    return {'tokenizer': {  'object': TextTilingTokenizer(w=w,
                                                          k=k,
                                                          similarity_method = similarity_method,
                                                          stopwords=stopwords,
                                                          smoothing_width=smoothing_width,
                                                          smoothing_rounds=smoothing_rounds,
                                                          cutoff_policy=cutoff_policy)}}




























